# Copyright 2016 Red Hat Inc
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import uuidutils
import random
import sqlalchemy as sa

from oslo_log import log as logging
from sqlalchemy import orm
from sqlalchemy.orm import exc as orm_exc
from tacker.db import db_base
from tacker.db import model_base
from tacker.db import models_v1
from tacker.db.nfvo.ns_db import NS
from tacker.db import types
from tacker.extensions import nfvo
from tacker.extensions.nfvo_plugins import vnffg
from tacker import manager
from tacker.plugins.common import constants


LOG = logging.getLogger(__name__)
_ACTIVE_UPDATE = (constants.ACTIVE, constants.PENDING_UPDATE)
_ACTIVE_UPDATE_ERROR_DEAD = (
    constants.PENDING_CREATE, constants.ACTIVE, constants.PENDING_UPDATE,
    constants.PENDING_DELETE, constants.ERROR, constants.DEAD)
_VALID_VNFFG_UPDATE_ATTRIBUTES = ('vnf_mapping',)
_VALID_SFC_UPDATE_ATTRIBUTES = ('chain', 'symmetrical')
_VALID_NFP_UPDATE_ATTRIBUTES = ('symmetrical',)
_VALID_FC_UPDATE_ATTRIBUTES = ()
MATCH_CRITERIA = (
    'eth_type', 'eth_src', 'eth_dst', 'vlan_id', 'vlan_pcp', 'mpls_label',
    'mpls_tc', 'ip_dscp', 'ip_ecn', 'ip_src_prefix', 'ip_dst_prefix',
    'ip_proto', 'destination_port_range', 'source_port_range',
    'network_src_port_id', 'network_dst_port_id', 'network_id', 'network_name',
    'tenant_id', 'icmpv4_type', 'icmpv4_code', 'arp_op', 'arp_spa',
    'arp_tpa', 'arp_sha', 'arp_tha', 'ipv6_src', 'ipv6_dst', 'ipv6_flabel',
    'icmpv6_type', 'icmpv6_code', 'ipv6_nd_target', 'ipv6_nd_sll',
    'ipv6_nd_tll')

MATCH_DB_KEY_LIST = (
    'eth_type', 'eth_src', 'eth_dst', 'vlan_id', 'vlan_pcp', 'mpls_label',
    'mpls_tc', 'ip_dscp', 'ip_ecn', 'ip_src_prefix', 'ip_dst_prefix',
    'ip_proto', 'destination_port_min', 'destination_port_max',
    'source_port_min', 'source_port_max', 'network_src_port_id',
    'network_dst_port_id', 'network_id', 'tenant_id', 'icmpv4_type',
    'icmpv4_code', 'arp_op', 'arp_spa', 'arp_tpa', 'arp_sha', 'arp_tha',
    'ipv6_src', 'ipv6_dst', 'ipv6_flabel', 'icmpv6_type', 'icmpv6_code',
    'ipv6_nd_target', 'ipv6_nd_sll', 'ipv6_nd_tll'
)

CP = 'connection_points'


class VnffgTemplate(model_base.BASE, models_v1.HasId, models_v1.HasTenant):
    """Represents template to create a VNF Forwarding Graph."""

    # Descriptive name
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.Text)

    # Vnffg template
    template = sa.Column(types.Json)

    # Vnffgd template source - onboarded
    template_source = sa.Column(sa.String(255), server_default='onboarded')


class Vnffg(model_base.BASE, models_v1.HasTenant, models_v1.HasId):
    """VNF Forwarding Graph Data Model"""

    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.String(255), nullable=True)

    # List of associated NFPs
    forwarding_paths = orm.relationship("VnffgNfp", backref="vnffg")

    vnffgd_id = sa.Column(types.Uuid, sa.ForeignKey('vnffgtemplates.id'))
    vnffgd = orm.relationship('VnffgTemplate')

    status = sa.Column(sa.String(255), nullable=False)

    # Mapping of VNFD to VNF instance names
    vnf_mapping = sa.Column(types.Json)

    attributes = sa.Column(types.Json)

    # Associated Network Service
    ns_id = sa.Column(types.Uuid, sa.ForeignKey('ns.id'), nullable=True)


class VnffgNfp(model_base.BASE, models_v1.HasTenant, models_v1.HasId):
    """Network Forwarding Path Data Model"""

    name = sa.Column(sa.String(255), nullable=False)
    vnffg_id = sa.Column(types.Uuid, sa.ForeignKey('vnffgs.id'),
                         nullable=False)

    # List of associated classifiers
    classifiers = orm.relationship('VnffgClassifier', backref='nfp')
    chain = orm.relationship('VnffgChain', backref='nfp',
                             uselist=False)

    status = sa.Column(sa.String(255), nullable=False)
    path_id = sa.Column(sa.String(255), nullable=True)

    # symmetry of forwarding path
    symmetrical = sa.Column(sa.Boolean(), default=False)


class VnffgChain(model_base.BASE, models_v1.HasTenant, models_v1.HasId):
    """Service Function Chain Data Model"""

    status = sa.Column(sa.String(255), nullable=False)

    instance_id = sa.Column(sa.String(255), nullable=True)

    # symmetry of forwarding path
    symmetrical = sa.Column(sa.Boolean(), default=False)

    # chain
    chain = sa.Column(types.Json)

    path_id = sa.Column(sa.String(255), nullable=True)
    nfp_id = sa.Column(types.Uuid, sa.ForeignKey('vnffgnfps.id'))


class VnffgClassifier(model_base.BASE, models_v1.HasTenant, models_v1.HasId):
    """VNFFG NFP Classifier Data Model"""

    name = sa.Column(sa.String(255), nullable=True)

    status = sa.Column(sa.String(255), nullable=False)

    instance_id = sa.Column(sa.String(255), nullable=True)

    chain_id = sa.Column(types.Uuid, sa.ForeignKey('vnffgchains.id'))
    chain = orm.relationship('VnffgChain', backref='classifier',
                             uselist=False, foreign_keys=[chain_id])
    nfp_id = sa.Column(types.Uuid, sa.ForeignKey('vnffgnfps.id'))
    # match criteria
    match = orm.relationship('ACLMatchCriteria')


class ACLMatchCriteria(model_base.BASE, models_v1.HasId):
    """Represents ACL match criteria of a classifier."""

    vnffgc_id = sa.Column(types.Uuid, sa.ForeignKey('vnffgclassifiers.id'))
    eth_src = sa.Column(sa.String(36), nullable=True)
    eth_dst = sa.Column(sa.String(36), nullable=True)
    eth_type = sa.Column(sa.String(36), nullable=True)
    vlan_id = sa.Column(sa.Integer, nullable=True)
    vlan_pcp = sa.Column(sa.Integer, nullable=True)
    mpls_label = sa.Column(sa.Integer, nullable=True)
    mpls_tc = sa.Column(sa.Integer, nullable=True)
    ip_dscp = sa.Column(sa.Integer, nullable=True)
    ip_ecn = sa.Column(sa.Integer, nullable=True)
    ip_src_prefix = sa.Column(sa.String(36), nullable=True)
    ip_dst_prefix = sa.Column(sa.String(36), nullable=True)
    source_port_min = sa.Column(sa.Integer, nullable=True)
    source_port_max = sa.Column(sa.Integer, nullable=True)
    destination_port_min = sa.Column(sa.Integer, nullable=True)
    destination_port_max = sa.Column(sa.Integer, nullable=True)
    ip_proto = sa.Column(sa.Integer, nullable=True)
    network_id = sa.Column(types.Uuid, nullable=True)
    network_src_port_id = sa.Column(types.Uuid, nullable=True)
    network_dst_port_id = sa.Column(types.Uuid, nullable=True)
    tenant_id = sa.Column(sa.String(64), nullable=True)
    icmpv4_type = sa.Column(sa.Integer, nullable=True)
    icmpv4_code = sa.Column(sa.Integer, nullable=True)
    arp_op = sa.Column(sa.Integer, nullable=True)
    arp_spa = sa.Column(sa.String(36), nullable=True)
    arp_tpa = sa.Column(sa.String(36), nullable=True)
    arp_sha = sa.Column(sa.String(36), nullable=True)
    arp_tha = sa.Column(sa.String(36), nullable=True)
    ipv6_src = sa.Column(sa.String(36), nullable=True)
    ipv6_dst = sa.Column(sa.String(36), nullable=True)
    ipv6_flabel = sa.Column(sa.Integer, nullable=True)
    icmpv6_type = sa.Column(sa.Integer, nullable=True)
    icmpv6_code = sa.Column(sa.Integer, nullable=True)
    ipv6_nd_target = sa.Column(sa.String(36), nullable=True)
    ipv6_nd_sll = sa.Column(sa.String(36), nullable=True)
    ipv6_nd_tll = sa.Column(sa.String(36), nullable=True)


class VnffgPluginDbMixin(vnffg.VNFFGPluginBase, db_base.CommonDbMixin):

    def __init__(self):
        super(VnffgPluginDbMixin, self).__init__()

    def create_vnffg(self, context, vnffg):
        vnffg_dict = self._create_vnffg_pre(context, vnffg)
        sfc_instance = uuidutils.generate_uuid()
        fc_instance = uuidutils.generate_uuid()
        self._create_vnffg_post(context, sfc_instance,
                                fc_instance, vnffg_dict)
        self._create_vnffg_status(context, vnffg_dict)
        return vnffg_dict

    def get_vnffg(self, context, vnffg_id, fields=None):
        vnffg_db = self._get_resource(context, Vnffg, vnffg_id)
        return self._make_vnffg_dict(vnffg_db, fields)

    def get_vnffgs(self, context, filters=None, fields=None):
        return self._get_collection(context, Vnffg, self._make_vnffg_dict,
                                    filters=filters, fields=fields)

    def update_vnffg(self, context, vnffg_id, vnffg):
        vnffg_dict = self._update_vnffg_pre(context, vnffg_id)
        self._update_vnffg_post(context, vnffg_id, constants.ACTIVE, vnffg)
        return vnffg_dict

    def delete_vnffg(self, context, vnffg_id):
        self._delete_vnffg_pre(context, vnffg_id)
        self._delete_vnffg_post(context, vnffg_id, False)

    def create_vnffgd(self, context, vnffgd):
        template = vnffgd['vnffgd']
        LOG.debug('template %s', template)
        tenant_id = self._get_tenant_id_for_create(context, template)
        template_source = template.get('template_source')

        with context.session.begin(subtransactions=True):
            template_id = uuidutils.generate_uuid()
            template_db = VnffgTemplate(
                id=template_id,
                tenant_id=tenant_id,
                name=template.get('name'),
                description=template.get('description'),
                template=template.get('template'),
                template_source=template_source)
            context.session.add(template_db)

        LOG.debug('template_db %(template_db)s',
                  {'template_db': template_db})
        return self._make_template_dict(template_db)

    def get_vnffgd(self, context, vnffgd_id, fields=None):
        template_db = self._get_resource(context, VnffgTemplate,
                                         vnffgd_id)
        return self._make_template_dict(template_db, fields)

    def get_vnffgds(self, context, filters=None, fields=None):
        if ('template_source' in filters) and \
                (filters['template_source'][0] == 'all'):
            filters.pop('template_source')
        return self._get_collection(context, VnffgTemplate,
                                    self._make_template_dict,
                                    filters=filters, fields=fields)

    def delete_vnffgd(self, context, vnffgd_id):
        with context.session.begin(subtransactions=True):
            vnffg_db = context.session.query(Vnffg).filter_by(
                vnffgd_id=vnffgd_id).first()
            if vnffg_db is not None:
                raise nfvo.VnffgdInUse(vnffgd_id=vnffgd_id)

            template_db = self._get_resource(context, VnffgTemplate,
                                             vnffgd_id)
            context.session.delete(template_db)

    def get_classifier(self, context, classifier_id, fields=None):
        classifier_db = self._get_resource(context, VnffgClassifier,
                                           classifier_id)
        return self._make_classifier_dict(classifier_db, fields)

    def get_classifiers(self, context, filters=None, fields=None):
        return self._get_collection(context, VnffgClassifier,
                                    self._make_classifier_dict,
                                    filters=filters, fields=fields)

    def create_classifiers_map(self, classifier_ids, instance_ids):
        return {classifier_id: instance_ids[i]
                for i, classifier_id in enumerate(classifier_ids)}

    def get_nfp(self, context, nfp_id, fields=None):
        nfp_db = self._get_resource(context, VnffgNfp, nfp_id)
        return self._make_nfp_dict(nfp_db, fields)

    def get_nfps(self, context, filters=None, fields=None):
        return self._get_collection(context, VnffgNfp,
                                    self._make_nfp_dict,
                                    filters=filters, fields=fields)

    def get_sfc(self, context, sfc_id, fields=None):
        chain_db = self._get_resource(context, VnffgChain, sfc_id)
        return self._make_chain_dict(chain_db, fields)

    def get_sfcs(self, context, filters=None, fields=None):
        return self._get_collection(context, VnffgChain,
                                    self._make_chain_dict,
                                    filters=filters, fields=fields)

    def _update_template_params(self, original, paramvalues, param_matched):
        if 'get_input' not in str(original):
            return
        if isinstance(original, dict):
            for key_, value in original.items():
                if isinstance(value, dict) and 'get_input' in value:
                    if value['get_input'] in paramvalues:
                        original[key_] = paramvalues[value['get_input']]
                        param_matched.setdefault(value['get_input'], 0)
                        param_matched[value['get_input']] += 1
                    else:
                        raise nfvo.VnffgTemplateParamParsingException(
                            get_input=value['get_input'])
                else:
                    self._update_template_params(value,
                                                 paramvalues, param_matched)
        elif isinstance(original, list):
            for element in original:
                self._update_template_params(element,
                                             paramvalues, param_matched)

    def _process_parameterized_template(self, dev_attrs, vnffgd_template):
        param_vattrs_dict = dev_attrs.pop('param_values', None)
        param_matched = {}
        if isinstance(param_vattrs_dict, dict):
            self._update_template_params(vnffgd_template,
                                param_vattrs_dict, param_matched)
        else:
            raise nfvo.VnffgParamValueFormatError(
                param_value=param_vattrs_dict)
        for param_key in param_vattrs_dict:
            if param_matched.get(param_key) is None:
                LOG.warning("Param input %s not used.", param_key)

    def _parametrize_topology_template(self, vnffg, template_db):
        if vnffg.get('attributes') and \
                vnffg['attributes'].get('param_values'):
            vnffg_param = vnffg['attributes']
            vnffgd_topology_template = \
                template_db.template['vnffgd']['topology_template']
            self._process_parameterized_template(vnffg_param,
                                                 vnffgd_topology_template)
            template_db.template['vnffgd']['topology_template'] = \
                vnffgd_topology_template

    # called internally, not by REST API
    def _create_vnffg_pre(self, context, vnffg):
        vnffg = vnffg['vnffg']
        LOG.debug('vnffg %s', vnffg)
        tenant_id = self._get_tenant_id_for_create(context, vnffg)
        name = vnffg.get('name')
        vnffg_id = vnffg.get('id') or uuidutils.generate_uuid()
        template_id = vnffg['vnffgd_id']
        ns_id = vnffg.get('ns_id', None)
        symmetrical_in_temp = self._get_symmetrical_template(context, vnffg)
        symmetrical = symmetrical_in_temp if symmetrical_in_temp is not None \
            else vnffg.get('symmetrical')

        with context.session.begin(subtransactions=True):
            template_db = self._get_resource(context, VnffgTemplate,
                                             template_id)
            LOG.debug('vnffg template %s', template_db)

            self._parametrize_topology_template(vnffg, template_db)

            vnf_members = self._get_vnffg_property(template_db.template,
                                                   'constituent_vnfs')
            LOG.debug('Constituent VNFs: %s', vnf_members)
            vnf_mapping = self._get_vnf_mapping(context, vnffg.get(
                                                'vnf_mapping'), vnf_members)
            LOG.debug('VNF Mapping: %s', vnf_mapping)
            # create NFP dict
            nfp_dict = self._create_nfp_pre(template_db)
            LOG.debug('NFP: %s', nfp_dict)
            path_id = nfp_dict['path_id']
            try:
                if path_id:
                    vnffgNfp_db = (self._model_query(context, VnffgNfp).
                                   filter(VnffgNfp.path_id == path_id).one())
                    raise nfvo.NfpDuplicatePathID(path_id=path_id,
                                                  nfp_name=vnffgNfp_db.name,
                                                  vnffg_name=name)
            except orm_exc.NoResultFound:
                pass

            vnffg_db = Vnffg(id=vnffg_id,
                             tenant_id=tenant_id,
                             name=name,
                             description=template_db.description,
                             vnf_mapping=vnf_mapping,
                             vnffgd_id=template_id,
                             ns_id=ns_id,
                             attributes=template_db.get('template'),
                             status=constants.PENDING_CREATE)
            context.session.add(vnffg_db)

            nfp_id = uuidutils.generate_uuid()
            sfc_id = uuidutils.generate_uuid()

            classifiers = self._policy_to_acl_criteria(context, template_db,
                                                   nfp_dict['name'],
                                                   vnf_mapping)
            LOG.debug('classifiers %s', classifiers)

            classifier_ids = [uuidutils.generate_uuid() for i in classifiers]

            nfp_db = VnffgNfp(id=nfp_id, vnffg_id=vnffg_id,
                              tenant_id=tenant_id,
                              name=nfp_dict['name'],
                              status=constants.PENDING_CREATE,
                              path_id=path_id,
                              symmetrical=symmetrical)
            context.session.add(nfp_db)

            chain = self._create_port_chain(context, vnf_mapping, template_db,
                                            nfp_dict['name'])
            LOG.debug('chain: %s', chain)
            sfc_db = VnffgChain(id=sfc_id,
                                tenant_id=tenant_id,
                                status=constants.PENDING_CREATE,
                                symmetrical=symmetrical,
                                chain=chain,
                                nfp_id=nfp_id,
                                path_id=path_id)

            context.session.add(sfc_db)

            for i, classifier_id in enumerate(classifier_ids):

                sfcc_db = VnffgClassifier(id=classifier_id,
                                          name=classifiers[i]['name'],
                                          tenant_id=tenant_id,
                                          status=constants.PENDING_CREATE,
                                          nfp_id=nfp_id,
                                          chain_id=sfc_id)
                context.session.add(sfcc_db)

                match_db_table = ACLMatchCriteria(
                    id=uuidutils.generate_uuid(),
                    vnffgc_id=classifier_id,
                    tenant_id=tenant_id,
                    **classifiers[i]['match'])

                context.session.add(match_db_table)

        return self._make_vnffg_dict(vnffg_db)

    @staticmethod
    def _create_nfp_pre(template_db):
        template = template_db.template['vnffgd']['topology_template']
        nfp_dict = dict()
        vnffg_name = list(template['groups'].keys())[0]
        # we assume only one NFP for initial implementation
        nfp_dict['name'] = template['groups'][vnffg_name]['members'][0]
        nfp_dict['path_id'] = template['node_templates'][nfp_dict['name']][
            'properties'].get('id', None)
        # 'path_id' will be updated when creating port chain is done
        return nfp_dict

    def _create_port_chain(self, context, vnf_mapping, template_db, nfp_name):
        """Creates a list of physical port ids to represent an ordered chain

        :param context: SQL session context
        :param vnf_mapping: dict of VNFD to VNF instance mappings
        :param template_db: VNFFG Descriptor
        :param nfp_name: name of the forwarding path with chain requirements
        :return: list of port chain including vnf name and list of CPs
        """
        chain_list = []
        prev_forwarder = None
        vnfm_plugin = manager.TackerManager.get_service_plugins()['VNFM']
        # Build the list of logical chain representation
        logical_chain = self._get_nfp_attribute(template_db.template,
                                                nfp_name, 'path')
        # Build physical port chain
        for element in logical_chain:
            if element['forwarder'] not in vnf_mapping:
                raise nfvo.NfpForwarderNotFoundException(vnfd=element[
                                                         'forwarder'],
                                                         mapping=vnf_mapping)
            # TODO(trozet): validate CP in VNFD has forwarding capability
            # Find VNF resources
            vnf = vnfm_plugin.get_vnf_resources(context,
                                                vnf_mapping[element[
                                                    'forwarder']]
                                                )
            vnf_info = vnfm_plugin.get_vnf(context,
                                           vnf_mapping[element['forwarder']])
            vnf_cp = None
            for resource in vnf:
                if resource['name'] == element['capability']:
                    vnf_cp = resource['id']
                    break
            if vnf_cp is None:
                raise nfvo.VnffgCpNotFoundException(cp_id=element[
                    'capability'], vnf_id=vnf_mapping[element['forwarder']])
            # Check if this is a new VNF entry in the chain
            if element['forwarder'] != prev_forwarder:
                chain_list.append(
                    {'name': vnf_info['name'],
                     CP: [vnf_cp],
                     'sfc_encap': element.get('sfc_encap', True)})
                prev_forwarder = element['forwarder']
            # Must be an egress CP
            else:
                if len(chain_list[-1][CP]) > 1:
                    raise nfvo.NfpRequirementsException(vnfd=element[
                        'forwarder'])
                else:
                    chain_list[-1][CP].append(vnf_cp)
        return chain_list

    @staticmethod
    def _get_vnffg_property(template, vnffg_property):
        template = template['vnffgd']['topology_template']
        vnffg_name = list(template['groups'].keys())[0]
        try:
            return template['groups'][vnffg_name]['properties'][vnffg_property]
        except KeyError:
            raise nfvo.VnffgPropertyNotFoundException(
                vnffg_property=vnffg_property)

    @staticmethod
    def _get_nfp_attribute(template, nfp, attribute):
        """Finds any attribute of an NFP described in a template

        :param template: VNFFGD template
        :param nfp: name of NFP
        :param attribute: attribute to find
        :return: value of attribute from template
        """
        template = template['vnffgd']['topology_template']
        try:
            attr_val = VnffgPluginDbMixin._search_value(
                template['node_templates'][nfp], attribute)
            if attr_val is None:
                LOG.debug('NFP %(nfp)s, attr %(attr)s',
                          {'nfp': template['node_templates'][nfp],
                           'attr': attribute})
                raise nfvo.NfpAttributeNotFoundException(attribute=attribute)
            else:
                return attr_val
        except KeyError:
            raise nfvo.NfpAttributeNotFoundException(attribute=attribute)

    @staticmethod
    def _search_value(search_dict, search_key):
        for k, v in search_dict.items():
            if k == search_key:
                return v
            elif isinstance(v, dict):
                val = VnffgPluginDbMixin._search_value(v, search_key)
                if val is not None:
                    return val

    def _validate_vnfd_in_vnf_mapping(self, vnf_mapping, vnf_members):
        """Validate whether or not the vnf_mapping is valid for update.

        In the update_vnnfg procedure we need to know whether or not the
        the vnf_mapping is valid so we can use it to update the chain.
        """
        if not vnf_mapping:
            raise nfvo.VnfMappingNotFoundException()
        else:
            for vnfd, vnf in vnf_mapping.items():
                if vnfd not in vnf_members:
                    raise nfvo.VnfMappingNotValidException(vnfd=vnfd)

    def _combine_current_and_new_vnf_mapping(self, context,
                                             new_mapping, old_mapping):
        """Create an updated vnf mapping.

        In this function we create an updated vnf mapping which is
        a mix of the vnf_mapping which already exists in database
        and the new mapping that the user passes.
        """
        updated_vnf_mapping = old_mapping.copy()
        updated_vnf_mapping.update(new_mapping)
        return updated_vnf_mapping

    def _get_vnf_mapping(self, context, vnf_mapping, vnf_members):
        """Creates/validates a mapping of VNFD names to VNF IDs for NFP.

        :param context: SQL session context
        :param vnf_mapping: dict of requested VNFD:VNF_ID mappings
        :param vnf_members: list of constituent VNFs from a VNFFG
        :return: dict of VNFD:VNF_ID mappings
        """
        vnfm_plugin = manager.TackerManager.get_service_plugins()['VNFM']
        new_mapping = dict()

        for vnfd in vnf_members:
            # there should only be one ID returned for a unique name
            try:
                vnfd_id = vnfm_plugin.get_vnfds(context, {'name': [vnfd]},
                                                fields=['id']).pop()['id']
            except Exception:
                raise nfvo.VnffgdVnfdNotFoundException(vnfd_name=vnfd)
            if vnfd_id is None:
                raise nfvo.VnffgdVnfdNotFoundException(vnfd_name=vnfd)
            else:
                # if no VNF mapping, we need to abstractly look for instances
                # that match VNFD
                if vnf_mapping is None or vnfd not in vnf_mapping:
                    # find suitable VNFs from vnfd_id
                    LOG.debug('Searching VNFS with id %s', vnfd_id)
                    vnf_list = vnfm_plugin.get_vnfs(context,
                                                    {'vnfd_id': [vnfd_id]},
                                                    fields=['id'])
                    if len(vnf_list) == 0:
                        raise nfvo.VnffgInvalidMappingException(vnfd_name=vnfd)
                    else:
                        LOG.debug('Matching VNFs found %s', vnf_list)
                        vnf_list = [vnf['id'] for vnf in vnf_list]
                    if len(vnf_list) > 1:
                        new_mapping[vnfd] = random.choice(vnf_list)
                    else:
                        new_mapping[vnfd] = vnf_list[0]
                # if VNF mapping, validate instances exist and match the VNFD
                else:
                    vnf_vnfd = vnfm_plugin.get_vnf(context, vnf_mapping[vnfd],
                                                   fields=['vnfd_id'])
                    if vnf_vnfd is not None:
                        vnf_vnfd_id = vnf_vnfd['vnfd_id']
                    else:
                        raise nfvo.VnffgInvalidMappingException(vnfd_name=vnfd)
                    if vnfd_id != vnf_vnfd_id:
                        raise nfvo.VnffgInvalidMappingException(vnfd_name=vnfd)
                    else:
                        new_mapping[vnfd] = vnf_mapping.pop(vnfd)
        self._validate_vim(context, new_mapping.values())
        return new_mapping

    def _validate_vim(self, context, vnfs):
        """Validates all VNFs are in the same VIM

        :param context: SQL Session Context
        :param vnfs: List of VNF instance IDs
        :return: None
        """
        LOG.debug('validating vim for vnfs %s', vnfs)
        vnfm_plugin = manager.TackerManager.get_service_plugins()['VNFM']
        vim_id = None
        for vnf in vnfs:
            vnf_dict = vnfm_plugin.get_vnf(context, vnf)
            if vim_id is None:
                vim_id = vnf_dict['vim_id']
            elif vnf_dict['vim_id'] != vim_id:
                raise nfvo.VnffgVimMappingException(vnf_id=vnf, vim_id=vim_id)

    def _validate_criteria(self, criteria):
        """Validate whether or not the classifiers are unique.

        We define a classifier as unique when at least one
        key-value pair is different from another classifier.
        """
        if not criteria:
            raise nfvo.NfpPolicyCriteriaIndexError()
        elif len(criteria) == 1:
            pass
        else:
            for index, dict_one in enumerate(criteria):
                if index != (len(criteria) - 1):
                    for dict_two in criteria[(index + 1):]:
                        if dict_one == dict_two:
                            raise nfvo. \
                                NfpDuplicatePolicyCriteria(first_dict=dict_one,
                                                           sec_dict=dict_two)

    def _policy_to_acl_criteria(self, context, template_db, nfp_name,
                                vnf_mapping):
        template = template_db.template['vnffgd']['topology_template']
        nfp = template['node_templates'][nfp_name]

        if 'policy' in nfp['properties']:
            policy = nfp['properties']['policy']
            if 'type' in policy:
                if policy['type'] != 'ACL':
                    raise nfvo.NfpPolicyTypeError(type=policy['type'])

            if 'criteria' not in policy:
                raise nfvo.NfpPolicyCriteriaError(
                    error="Missing criteria in policy")
            validation_list = []
            for item in policy['criteria']:
                if item.get('name') is None:
                    LOG.warning('The unnamed classifier approach'
                                ' will be deprecated in subsequent'
                                ' releases')
                    validation_list.append(item)
                else:
                    validation_list.append(item['classifier'])

            self._validate_criteria(validation_list)

            classifiers = []
            for criteria in policy['criteria']:
                match = dict()
                if criteria.get('name') is None:
                    criteria_dict = criteria.copy()
                else:
                    criteria_dict = criteria['classifier'].copy()
                for key, val in criteria_dict.items():
                    if key in MATCH_CRITERIA:
                        match.update(self._convert_criteria(context, key, val,
                                                            vnf_mapping))
                    else:
                        raise nfvo.NfpPolicyCriteriaError(error="Unsupported "
                                                          "criteria: "
                                                          "{}".format(key))
                classifiers.append({'name': criteria.get('name'),
                                    'match': match})
            return classifiers
        else:
            return []

    def _convert_criteria(self, context, criteria, value, vnf_mapping):
        """Method is used to convert criteria to proper db value from template

        :param context: SQL session context
        :param criteria: input criteria name
        :param value: input value
        :param vnf_mapping: mapping of VNFD to VNF instances
        :return: converted dictionary
        """

        if criteria.endswith('_range'):
            prefix = criteria[:-6]
            criteria_min = prefix + "_min"
            criteria_max = prefix + "_max"
            try:
                min_val, max_val = value.split('-')
            except ValueError:
                raise nfvo.NfpPolicyCriteriaError(error="Range missing or "
                                                  "incorrect for "
                                                  "{}".format(criteria))
            return {criteria_min: int(min_val), criteria_max: int(max_val)}

        elif criteria.endswith('_name'):
            prefix = criteria[:-5]
            vnf_id = list(vnf_mapping.values())[0]
            new_value = self._vim_resource_name_to_id(context, prefix, value,
                                                      vnf_id)
            new_name = prefix + "_id"
            return {new_name: new_value}

        else:
            return {criteria: value}

    def _vim_resource_name_to_id(self, context, resource, name, vnf_id):
        """Converts a VIM resource name to its ID

        :param context: SQL session context
        :param resource: resource type to find (network, subnet, etc)
        :param name: name of the resource to find its ID
        :param vnf_id: A VNF instance ID that is part of the chain to which
               the classifier will apply to
        :return: ID of the resource name
        """
        # this should be overridden with driver call to find ID given name
        # for resource
        return uuidutils.generate_uuid()

    # called internally, not by REST API
    # instance_id = None means error on creation
    def _create_vnffg_post(self, context, sfc_instance_id, path_id,
                           classifiers_map, vnffg_dict):
        LOG.debug('SFC created instance is %s', sfc_instance_id)
        LOG.debug('Flow Classifiers created instances are %s',
                  [classifiers_map[item] for item in classifiers_map])
        nfp_dict = self.get_nfp(context, vnffg_dict['forwarding_paths'])
        sfc_id = nfp_dict['chain_id']
        with context.session.begin(subtransactions=True):
            nfp_query = (self._model_query(context, VnffgNfp).
                         filter(VnffgNfp.id == nfp_dict['id']).
                         filter(VnffgNfp.status == constants.PENDING_CREATE).
                         one())
            nfp_query.update({'path_id': path_id})
            query = (self._model_query(context, VnffgChain).
                     filter(VnffgChain.id == sfc_id).
                     filter(VnffgChain.status == constants.PENDING_CREATE).
                     one())
            query.update({'instance_id': sfc_instance_id, 'path_id': path_id})
            if sfc_instance_id is None:
                query.update({'status': constants.ERROR})
            else:
                query.update({'status': constants.ACTIVE})
            for classifier_id, fc_instance_id in classifiers_map.items():
                query = (self._model_query(context, VnffgClassifier).
                         filter(VnffgClassifier.id == classifier_id).
                         filter(VnffgClassifier.status ==
                                constants.PENDING_CREATE).
                         one())
                query.update({'instance_id': fc_instance_id})

                if fc_instance_id is None:
                    query.update({'status': constants.ERROR})
                else:
                    query.update({'status': constants.ACTIVE})

    def _create_vnffg_status(self, context, vnffg):
        nfp = self.get_nfp(context, vnffg['forwarding_paths'])
        chain = self.get_sfc(context, nfp['chain_id'])

        if chain['status'] == constants.ERROR:
            self._update_all_status(context, vnffg['id'], nfp['id'],
                                    constants.ERROR)

        elif chain['status'] == constants.ACTIVE:
            classifiers_active_state = True
            for classifier in [self.get_classifier(context, classifier_id)
                               for classifier_id in nfp['classifier_ids']]:

                if classifier['status'] == constants.ACTIVE:
                    continue
                elif classifier['status'] == constants.ERROR:
                    classifiers_active_state = False
                    break

            if classifiers_active_state:
                self._update_all_status(context, vnffg['id'], nfp['id'],
                                    constants.ACTIVE)
            else:
                self._update_all_status(context, vnffg['id'], nfp['id'],
                                    constants.ERROR)

    def _update_all_status(self, context, vnffg_id, nfp_id, status):
        nfp_dict = self.get_nfp(context, nfp_id)
        sfc_id = nfp_dict['chain_id']

        with context.session.begin(subtransactions=True):
            for classifier_id in nfp_dict['classifier_ids']:
                query_cls = (self._model_query(context, VnffgClassifier).
                             filter(VnffgClassifier.id == classifier_id))
                query_cls.update({'status': status})
            query_chain = (self._model_query(context, VnffgChain).
                           filter(VnffgChain.id == sfc_id))
            query_chain.update({'status': status})
            query = (self._model_query(context, Vnffg).
                     filter(Vnffg.id == vnffg_id))
            query.update({'status': status})
            nfp_query = (self._model_query(context, VnffgNfp).
                         filter(VnffgNfp.id == nfp_id))
            nfp_query.update({'status': status})

    def _make_vnffg_dict(self, vnffg_db, fields=None):
        LOG.debug('vnffg_db %s', vnffg_db)
        LOG.debug('vnffg_db nfp %s', vnffg_db.forwarding_paths)
        res = {
            'forwarding_paths': vnffg_db.forwarding_paths[0]['id']
        }
        key_list = ('id', 'tenant_id', 'name', 'description', 'ns_id',
                    'vnf_mapping', 'status', 'vnffgd_id', 'attributes')
        res.update((key, vnffg_db[key]) for key in key_list)
        return self._fields(res, fields)

    def _update_vnffg_status_pre(self, context, vnffg_id):
        vnffg = self.get_vnffg(context, vnffg_id)
        nfp = self.get_nfp(context, vnffg['forwarding_paths'])
        sfc = self.get_sfc(context, nfp['chain_id'])
        classifiers = [self.get_classifier(context, classifier_id) for
                       classifier_id in nfp['classifier_ids']]
        with context.session.begin(subtransactions=True):
            vnffg_db = self._get_vnffg_db(context, vnffg['id'], _ACTIVE_UPDATE,
                                          constants.PENDING_UPDATE)
            self._get_nfp_db(context, nfp['id'], _ACTIVE_UPDATE,
                             constants.PENDING_UPDATE)
            self._get_sfc_db(context, sfc['id'], _ACTIVE_UPDATE,
                             constants.PENDING_UPDATE)
            for classifier in classifiers:
                self._get_classifier_db(context, classifier['id'],
                                        _ACTIVE_UPDATE,
                                        constants.PENDING_UPDATE)
        return self._make_vnffg_dict(vnffg_db)

    def _update_vnffg_pre(self, context, vnffg, vnffg_id, vnffg_old):
        vnffg = vnffg['vnffg']
        del vnffg['symmetrical']
        if vnffg.get('vnffgd_template') is None:
            try:
                return self._update_vnffg_without_template(context, vnffg_old,
                                                           vnffg, vnffg_id)
            except (nfvo.VnfMappingNotFoundException,
                    nfvo.VnfMappingNotValidException) as e:
                raise e

        with context.session.begin(subtransactions=True):
            # Templates
            template_db_new = self._get_resource(context, VnffgTemplate,
                                                 vnffg['vnffgd_id'])

            LOG.debug('vnffg new template %s', template_db_new)

            template_db_old = self._get_resource(context, VnffgTemplate,
                                                 vnffg_old['vnffgd_id'])

            LOG.debug('vnffg old template %s', template_db_old)

            self._parametrize_topology_template(vnffg, template_db_new)

            # VNF-Members
            vnf_members_new = self._get_vnffg_property(
                template_db_new.template, 'constituent_vnfs')

            LOG.debug('New Constituent VNFs: %s', vnf_members_new)

            vnf_members_old = self._get_vnffg_property(
                template_db_old.template, 'constituent_vnfs')

            LOG.debug('Old Constituent VNFs: %s', vnf_members_old)

            if set(vnf_members_new) == set(vnf_members_old):
                if vnffg.get('vnf_mapping') is None:
                    final_vnf_mapping = vnffg_old['vnf_mapping']
                else:
                    try:
                        self._validate_vnfd_in_vnf_mapping(
                            vnffg['vnf_mapping'], vnf_members_new)
                    except (nfvo.VnfMappingNotFoundException,
                            nfvo.VnfMappingNotValidException) as e:
                        raise e
                    updated_vnf_mapping = \
                        self._combine_current_and_new_vnf_mapping(
                            context, vnffg['vnf_mapping'],
                            vnffg_old['vnf_mapping'])

                    final_vnf_mapping = self._get_vnf_mapping(
                        context, updated_vnf_mapping, vnf_members_new)
            else:
                final_vnf_mapping = self._get_vnf_mapping(context, vnffg.get(
                                                          'vnf_mapping'),
                                                          vnf_members_new)

            LOG.debug('VNF Mapping: %s', final_vnf_mapping)
            # Update the vnffg with the new template.
            query_vnffg = (self._model_query(context, Vnffg).
                           filter(Vnffg.id == vnffg_old['id']).
                           filter(Vnffg.status == constants.PENDING_UPDATE))
            query_vnffg.update({'vnf_mapping': final_vnf_mapping,
                                'vnffgd_id': vnffg['vnffgd_id'],
                                'description': template_db_new.description,
                                'attributes': template_db_new.get('template')})

            # Delete the old_vnffgd_template if template_source is 'inline'
            if template_db_old.template_source == 'inline':
                self.delete_vnffgd(context, vnffg_old['vnffgd_id'])

            # update NFP
            nfp_dict_old = self.get_nfp(context, vnffg_old['forwarding_paths'])

            LOG.debug('Current NFP: %s', nfp_dict_old)

            nfp_dict_new = self._update_nfp_pre(template_db_new, nfp_dict_old)

            LOG.debug('New NFP: %s', nfp_dict_new)

            query_nfp = (self._model_query(context, VnffgNfp).
                     filter(VnffgNfp.id == nfp_dict_old['id']).
                     filter(VnffgNfp.status == constants.PENDING_UPDATE))
            query_nfp.update(nfp_dict_new)

            # update chain
            chain_old = self.get_sfc(context, nfp_dict_old['chain_id'])
            LOG.debug('Current chain: %s', chain_old)
            chain_new = self._create_port_chain(context, final_vnf_mapping,
                                                template_db_new,
                                                nfp_dict_new['name'])
            LOG.debug('New chain: %s', chain_new)
            # to check if it is updated
            update_chain = self._set_updated_chain(chain_old['chain'],
                                                   chain_new)

            if update_chain:
                query_chain = (self._model_query(context, VnffgChain).
                               filter(VnffgChain.id == chain_old['id']).
                               filter(VnffgChain.status == constants.
                               PENDING_UPDATE))
                query_chain.update({'chain': chain_new,
                                    'path_id': nfp_dict_new['path_id']})

            # update classifiers
            classifiers_old = []
            for classifier_id in nfp_dict_old['classifier_ids']:
                classifiers_old.append(self.
                                   get_classifier(context,
                                   classifier_id,
                                   fields=['name', 'match', 'id']))
            classifiers_new = self._policy_to_acl_criteria(context,
                                                       template_db_new,
                                                       nfp_dict_new['name'],
                                                       final_vnf_mapping)
            try:
                classifiers_update, classifiers_delete = \
                    self._find_classifiers_to_update(classifiers_old,
                                                     classifiers_new)
            except nfvo.UpdateVnffgException as e:
                raise e
            for clsfr in classifiers_update:
                if clsfr.get('id'):
                    for item in MATCH_DB_KEY_LIST:
                        if clsfr['match'].get(item) is None:
                            clsfr['match'][item] = None
                    query_match = (self._model_query(context,
                                                     ACLMatchCriteria).
                             filter(ACLMatchCriteria.vnffgc_id == clsfr['id']))
                    query_match.update(clsfr['match'])
                else:
                    classifier_id = uuidutils.generate_uuid()
                    sfcc_db = VnffgClassifier(id=classifier_id,
                                          name=clsfr['name'],
                                          tenant_id=vnffg_old['tenant_id'],
                                          status=constants.PENDING_CREATE,
                                          nfp_id=nfp_dict_old['id'],
                                          chain_id=chain_old['id'])
                    context.session.add(sfcc_db)

                    match_db = ACLMatchCriteria(
                        id=uuidutils.generate_uuid(),
                        vnffgc_id=classifier_id,
                        **clsfr['match'])
                    context.session.add(match_db)
            for clsfr in classifiers_delete:
                query_clsfr = (self._model_query(context, VnffgClassifier).
                               filter(VnffgClassifier.id == clsfr['id']).
                               filter(VnffgClassifier.status == constants.
                                     PENDING_UPDATE))
                query_clsfr.update({'status': constants.PENDING_DELETE})

        return self.get_vnffg(context, vnffg_id)

    def _find_classifiers_to_update(self, current_classifiers,
                                    new_classifiers):
        update_classifiers = []
        delete_classifiers = []
        names_list = []
        for new_clsfr in new_classifiers:
            found_name = False
            if new_clsfr['name'] is None:
                LOG.error('VNFFG update requires named classifiers')
                raise nfvo.UpdateVnffgException(
                    message="Failed to update VNFFG")
            for cur_clsfr in current_classifiers:
                if cur_clsfr['name'] == new_clsfr['name']:
                    new_clsfr['id'] = cur_clsfr['id']
                    names_list.append(new_clsfr['name'])
                    update_classifiers.append(new_clsfr)
                    found_name = True
                    break
            if not found_name:
                names_list.append(new_clsfr['name'])
                update_classifiers.append(new_clsfr)
        for cur_clsfr in current_classifiers:
            if cur_clsfr['name'] not in names_list:
                delete_classifiers.append(cur_clsfr)
        return update_classifiers, delete_classifiers

    def _set_updated_chain(self, current_chain, new_chain):
        if len(current_chain) != len(new_chain):
            return True
        else:
            for i, item in enumerate(current_chain):
                cp_vnf = new_chain[i]
                if (cp_vnf['name'] == item['name'] and
                        cp_vnf['connection_points'] == item[
                        'connection_points']):
                    continue
                else:
                    return True
        return False

    def _update_vnffg_without_template(self, context, old_vnffg, new_vnffg,
                                       vnffg_id):

        template_db = self._get_resource(context, VnffgTemplate,
                                         old_vnffg['vnffgd_id'])
        vnfd_members = self._get_vnffg_property(template_db.template,
                                                'constituent_vnfs')
        nfp = self.get_nfp(context, old_vnffg['forwarding_paths'])

        chain_dict = self.get_sfc(context, nfp['chain_id'])
        try:
            self._validate_vnfd_in_vnf_mapping(new_vnffg.get('vnf_mapping'),
                                               vnfd_members)
        except (nfvo.VnfMappingNotFoundException,
                nfvo.VnfMappingNotValidException) as e:
            raise e

        combined_vnf_mapping = self._combine_current_and_new_vnf_mapping(
            context, new_vnffg['vnf_mapping'], old_vnffg['vnf_mapping'])

        new_vnffg['vnf_mapping'] = self._get_vnf_mapping(context,
                                                         combined_vnf_mapping,
                                                         vnfd_members)
        new_chain = self._create_port_chain(context,
                                            new_vnffg['vnf_mapping'],
                                            template_db,
                                            nfp['name'])

        LOG.debug('chain update: %s', new_chain)

        query_vnffg = (self._model_query(context, Vnffg).
                       filter(Vnffg.id == old_vnffg['id']).
                       filter(Vnffg.status == constants.PENDING_UPDATE))
        query_vnffg.update({'vnf_mapping': new_vnffg['vnf_mapping']})

        query_chain = (self._model_query(context, VnffgChain).
                       filter(VnffgChain.id == chain_dict['id']).
                       filter(VnffgChain.status == constants.
                       PENDING_UPDATE))
        query_chain.update({'chain': new_chain})

        return self.get_vnffg(context, vnffg_id)

    def _update_nfp_pre(self, template_db, nfp_dict_old):
        template_new = template_db.template['vnffgd']['topology_template']
        nfp_dict_new = dict()
        vnffg_name = list(template_new['groups'].keys())[0]
        nfp_dict_new['name'] = template_new['groups'][vnffg_name]['members'][0]
        nfp_dict_new['path_id'] = template_new['node_templates'][nfp_dict_new[
            'name']]['properties'].get('id')

        if not nfp_dict_new['path_id']:
            nfp_dict_new['path_id'] = nfp_dict_old['path_id']
        return nfp_dict_new

    def _update_vnffg_post(self, context, n_sfc_chain_id,
                           classifiers_map, vnffg_dict):
        """Updates the status and the n-sfc instance_ids in the db

        :param context: SQL Session Context
        :param n_sfc_chain_id: Id of port-chain in n-sfc side
        :param classifiers_map: classifier and instance Ids map
        :param vnffg_dict: vnffg dictionary
        :return: None
        """
        nfp_dict = self.get_nfp(context, vnffg_dict['forwarding_paths'])
        sfc_id = nfp_dict['chain_id']
        with context.session.begin(subtransactions=True):
            query_chain = (self._model_query(context, VnffgChain).
                filter(VnffgChain.id == sfc_id).
                filter(VnffgChain.status == constants.PENDING_UPDATE).one())
            if n_sfc_chain_id is None:
                query_chain.update({'status': constants.ERROR})
            else:
                query_chain.update({'status': constants.ACTIVE})
            for clsfr_id in nfp_dict['classifier_ids']:
                query_clsfr = (self._model_query(context, VnffgClassifier).
                    filter(VnffgClassifier.id == clsfr_id))
                if classifiers_map.get(clsfr_id):
                    query_clsfr.update({
                        'instance_id': classifiers_map[clsfr_id]})
                    if classifiers_map[clsfr_id]:
                        query_clsfr.update({'status': constants.ACTIVE})
                    else:
                        query_clsfr.update({'status': constants.ERROR})
                else:
                    # Deletion of unused Match criterias which are
                    # not longer required due to the update classifier
                    # procedure.
                    query_match = (
                        self._model_query(context, ACLMatchCriteria).
                        filter(ACLMatchCriteria.vnffgc_id == clsfr_id))
                    query_match.delete()
                    query_clsfr.delete()

    def _update_vnffg_status_post(self, context, vnffg, error=False,
                                  db_state=constants.ERROR):

        nfp = self.get_nfp(context, vnffg['forwarding_paths'])
        chain = self.get_sfc(context, nfp['chain_id'])

        if error:
            if db_state == constants.ACTIVE:
                self._update_all_status(context, vnffg['id'], nfp['id'],
                                        constants.ACTIVE)
            else:
                self._update_all_status(context, vnffg['id'], nfp['id'],
                                        constants.ERROR)
        else:
            if chain['status'] == constants.ERROR:
                self._update_all_status(context, vnffg['id'], nfp['id'],
                                    constants.ERROR)
            elif chain['status'] == constants.ACTIVE:
                classifiers_active_state = True
                for classifier in [self.get_classifier(context, classifier_id)
                                   for classifier_id in nfp['classifier_ids']]:
                    if classifier['status'] == constants.ACTIVE:
                        continue
                    elif classifier['status'] == constants.ERROR:
                        classifiers_active_state = False
                        break
                if classifiers_active_state:
                    self._update_all_status(context, vnffg['id'], nfp['id'],
                                        constants.ACTIVE)
                else:
                    self._update_all_status(context, vnffg['id'], nfp['id'],
                                        constants.ERROR)

    def _get_vnffg_db(self, context, vnffg_id, current_statuses, new_status):
        try:
            vnffg_db = (
                self._model_query(context, Vnffg).
                filter(Vnffg.id == vnffg_id).
                filter(Vnffg.status.in_(current_statuses)).
                with_for_update().one())
        except orm_exc.NoResultFound:
            raise nfvo.VnffgNotFoundException(vnffg_id=vnffg_id)
        if vnffg_db.status == constants.PENDING_UPDATE:
            raise nfvo.VnffgInUse(vnffg_id=vnffg_id)
        vnffg_db.update({'status': new_status})
        return vnffg_db

    def _get_nfp_db(self, context, nfp_id, current_statuses, new_status):
        try:
            nfp_db = (
                self._model_query(context, VnffgNfp).
                filter(VnffgNfp.id == nfp_id).
                filter(VnffgNfp.status.in_(current_statuses)).
                with_for_update().one())
        except orm_exc.NoResultFound:
            raise nfvo.NfpNotFoundException(nfp_id=nfp_id)
        if nfp_db.status == constants.PENDING_UPDATE:
            raise nfvo.NfpInUse(nfp_id=nfp_id)
        nfp_db.update({'status': new_status})
        return nfp_db

    def _get_sfc_db(self, context, sfc_id, current_statuses, new_status):
        try:
            sfc_db = (
                self._model_query(context, VnffgChain).
                filter(VnffgChain.id == sfc_id).
                filter(VnffgChain.status.in_(current_statuses)).
                with_for_update().one())
        except orm_exc.NoResultFound:
            raise nfvo.SfcNotFoundException(sfc_id=sfc_id)
        if sfc_db.status == constants.PENDING_UPDATE:
            raise nfvo.SfcInUse(sfc_id=sfc_id)
        sfc_db.update({'status': new_status})
        return sfc_db

    def _get_classifier_db(self, context, fc_id, current_statuses, new_status):
        try:
            fc_db = (
                self._model_query(context, VnffgClassifier).
                filter(VnffgClassifier.id == fc_id).
                filter(VnffgClassifier.status.in_(current_statuses)).
                with_for_update().one())
        except orm_exc.NoResultFound:
            raise nfvo.ClassifierNotFoundException(fc_id=fc_id)
        if fc_db.status == constants.PENDING_UPDATE:
            raise nfvo.ClassifierInUse(fc_id=fc_id)
        fc_db.update({'status': new_status})
        return fc_db

    def _delete_vnffg_pre(self, context, vnffg_id):
        vnffg = self.get_vnffg(context, vnffg_id)
        ns_id = vnffg.get('ns_id')
        if ns_id:
            ns_db = self._get_resource(context, NS, ns_id)
            # If network service is not in pending_delete status,
            # raise error when delete vnffg.
            if ns_db['status'] != constants.PENDING_DELETE:
                raise nfvo.VnffgInUseNS(vnffg_id=vnffg_id,
                                        ns_id=vnffg.get('ns_id'))
        nfp = self.get_nfp(context, vnffg['forwarding_paths'])
        chain = self.get_sfc(context, nfp['chain_id'])
        classifiers = [self.get_classifier(context, classifier_id)
                       for classifier_id in nfp['classifier_ids']]
        with context.session.begin(subtransactions=True):
            vnffg_db = self._get_vnffg_db(
                context, vnffg['id'], _ACTIVE_UPDATE_ERROR_DEAD,
                constants.PENDING_DELETE)
            self._get_nfp_db(context, nfp['id'], _ACTIVE_UPDATE_ERROR_DEAD,
                             constants.PENDING_DELETE)
            self._get_sfc_db(context, chain['id'], _ACTIVE_UPDATE_ERROR_DEAD,
                             constants.PENDING_DELETE)
            for classifier in classifiers:
                self._get_classifier_db(context, classifier['id'],
                                        _ACTIVE_UPDATE_ERROR_DEAD,
                                        constants.PENDING_DELETE)

        return self._make_vnffg_dict(vnffg_db)

    def _delete_vnffg_post(self, context, vnffg_id, error):
        vnffg = self.get_vnffg(context, vnffg_id)
        nfp = self.get_nfp(context, vnffg['forwarding_paths'])
        chain = self.get_sfc(context, nfp['chain_id'])
        classifiers = [self.get_classifier(context, classifier_id)
                       for classifier_id in nfp['classifier_ids']]
        fc_queries = []
        match_queries = []
        with context.session.begin(subtransactions=True):
            vnffg_query = (
                self._model_query(context, Vnffg).
                filter(Vnffg.id == vnffg['id']).
                filter(Vnffg.status == constants.PENDING_DELETE))
            nfp_query = (
                self._model_query(context, VnffgNfp).
                filter(VnffgNfp.id == nfp['id']).
                filter(VnffgNfp.status == constants.PENDING_DELETE))
            sfc_query = (
                self._model_query(context, VnffgChain).
                filter(VnffgChain.id == chain['id']).
                filter(VnffgChain.status == constants.PENDING_DELETE))
            for classifier in classifiers:
                fc_queries.append((
                    self._model_query(context, VnffgClassifier).
                    filter(VnffgClassifier.id == classifier['id']).
                    filter(VnffgClassifier.status ==
                           constants.PENDING_DELETE)))
                match_queries.append((
                    self._model_query(context, ACLMatchCriteria).
                    filter(ACLMatchCriteria.vnffgc_id == classifier['id'])))
            if error:
                vnffg_query.update({'status': constants.ERROR})
                nfp_query.update({'status': constants.ERROR})
                sfc_query.update({'status': constants.ERROR})
                for fc_query in fc_queries:
                    fc_query.update({'status': constants.ERROR})
            else:
                for match_query in match_queries:
                    match_query.delete()
                for fc_query in fc_queries:
                    fc_query.delete()
                sfc_query.delete()
                nfp_query.delete()
                vnffg_query.delete()

            vnffgd_id = vnffg.get('vnffgd_id')
            template_db = self._get_resource(context, VnffgTemplate,
                                             vnffgd_id)

            if template_db.get('template_source') == 'inline':
                self.delete_vnffgd(context, vnffgd_id)

    def _get_symmetrical_template(self, context, vnffg):
        fp_prop = self._get_fp_properties(context, vnffg)
        return fp_prop.get('symmetrical', False)

    def _get_correlation_template(self, context, vnffg):
        fp_prop = self._get_fp_properties(context, vnffg)
        return fp_prop.get('correlation', 'mpls')

    def _get_fp_properties(self, context, vnffg):
        vnffgd_topo = None
        if vnffg.get('vnffgd_template'):
            vnffgd_topo = vnffg['vnffgd_template']['topology_template']
        elif vnffg.get('vnffgd_id'):
            vnffgd_template = self.get_vnffgd(context, vnffg.get('vnffgd_id'))
            vnffgd_topo = vnffgd_template['template']['vnffgd'][
                'topology_template']
        vnffg_name = list(vnffgd_topo['groups'].keys())[0]
        nfp_name = vnffgd_topo['groups'][vnffg_name]['members'][0]
        fp_prop = vnffgd_topo['node_templates'][nfp_name]['properties']
        return fp_prop

    def _make_template_dict(self, template, fields=None):
        res = {}
        key_list = ('id', 'tenant_id', 'name', 'description', 'template',
                    'template_source')
        res.update((key, template[key]) for key in key_list)
        return self._fields(res, fields)

    def _make_acl_match_dict(self, acl_match_db):
        key_list = MATCH_DB_KEY_LIST
        return {key: entry[key] for key in key_list for entry in acl_match_db
                if entry[key]}

    def _make_classifier_dict(self, classifier_db, fields=None):
        LOG.debug('classifier_db %s', classifier_db)
        LOG.debug('classifier_db match %s', classifier_db.match)
        res = {
            'match': self._make_acl_match_dict(classifier_db.match)
        }
        key_list = ('id', 'name', 'tenant_id', 'instance_id', 'status',
                    'chain_id', 'nfp_id')
        res.update((key, classifier_db[key]) for key in key_list)
        return self._fields(res, fields)

    def _make_nfp_dict(self, nfp_db, fields=None):
        LOG.debug('nfp_db %s', nfp_db)
        res = {'chain_id': nfp_db.chain['id'],
               'classifier_ids': [classifier['id'] for classifier in
                               nfp_db.classifiers]}
        key_list = ('name', 'id', 'tenant_id', 'symmetrical', 'status',
                    'path_id', 'vnffg_id')
        res.update((key, nfp_db[key]) for key in key_list)
        return self._fields(res, fields)

    def _make_chain_dict(self, chain_db, fields=None):
        LOG.debug('chain_db %s', chain_db)
        res = {}
        key_list = ('id', 'tenant_id', 'symmetrical', 'status', 'chain',
                    'path_id', 'nfp_id', 'instance_id')
        res.update((key, chain_db[key]) for key in key_list)
        return self._fields(res, fields)

    def _get_resource(self, context, model, res_id):
        try:
            return self._get_by_id(context, model, res_id)
        except orm_exc.NoResultFound:
            if issubclass(model, Vnffg):
                raise nfvo.VnffgNotFoundException(vnffg_id=res_id)
            elif issubclass(model, VnffgClassifier):
                raise nfvo.ClassifierNotFoundException(classifier_id=res_id)
            if issubclass(model, VnffgTemplate):
                raise nfvo.VnffgdNotFoundException(vnffgd_id=res_id)
            if issubclass(model, VnffgChain):
                raise nfvo.SfcNotFoundException(sfc_id=res_id)
            else:
                raise



# Copyright 2018 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_versionedobjects import fields


# Import fields from oslo.versionedobjects
StringField = fields.StringField
ListOfObjectsField = fields.ListOfObjectsField
ListOfStringsField = fields.ListOfStringsField
DictOfStringsField = fields.DictOfStringsField
DictOfNullableStringsField = fields.DictOfNullableStringsField
DateTimeField = fields.DateTimeField
BooleanField = fields.BooleanField
BaseEnumField = fields.BaseEnumField
Enum = fields.Enum
ObjectField = fields.ObjectField
IntegerField = fields.IntegerField
FieldType = fields.FieldType


class BaseTackerEnum(Enum):
    def __init__(self):
        super(BaseTackerEnum, self).__init__(valid_values=self.__class__.ALL)


class ContainerFormat(BaseTackerEnum):
    AKI = 'AKI'
    AMI = 'AMI'
    ARI = 'ARI'
    BARE = 'BARE'
    DOCKER = 'DOCKER'
    OVA = 'OVA'
    OVF = 'OVF'

    ALL = (AKI, AMI, ARI, BARE, DOCKER, OVA, OVF)


class ContainerFormatFields(BaseEnumField):
    AUTO_TYPE = ContainerFormat()


class DiskFormat(BaseTackerEnum):
    AKI = 'AKI'
    AMI = 'AMI'
    ARI = 'ARI'
    ISO = 'ISO'
    QCOW2 = 'QCOW2'
    RAW = 'RAW'
    VDI = 'VDI'
    VHD = 'VHD'
    VHDX = 'VHDX'
    VMDK = 'VMDK'

    ALL = (AKI, AMI, ARI, ISO, QCOW2, RAW, VDI, VHD, VHDX, VMDK)


class DiskFormatFields(BaseEnumField):
    AUTO_TYPE = DiskFormat()


class PackageOnboardingStateType(BaseTackerEnum):
    CREATED = 'CREATED'
    UPLOADING = 'UPLOADING'
    PROCESSING = 'PROCESSING'
    ONBOARDED = 'ONBOARDED'

    ALL = (CREATED, UPLOADING, PROCESSING, ONBOARDED)


class PackageOnboardingStateTypeField(BaseEnumField):
    AUTO_TYPE = PackageOnboardingStateType()


class PackageOperationalStateType(BaseTackerEnum):
    ENABLED = 'ENABLED'
    DISABLED = 'DISABLED'

    ALL = (ENABLED, DISABLED)


class PackageOperationalStateTypeField(BaseEnumField):
    AUTO_TYPE = PackageOperationalStateType()


class PackageUsageStateType(BaseTackerEnum):
    IN_USE = 'IN_USE'
    NOT_IN_USE = 'NOT_IN_USE'

    ALL = (IN_USE, NOT_IN_USE)


class PackageUsageStateTypeField(BaseEnumField):
    AUTO_TYPE = PackageUsageStateType()


class DictOfNullableField(fields.AutoTypedField):
    AUTO_TYPE = fields.Dict(fields.FieldType(), nullable=True)


class UUID(fields.UUID):
    def coerce(self, obj, attr, value):
        uuid.UUID(str(value))
        return str(value)


class UUIDField(fields.AutoTypedField):
    AUTO_TYPE = UUID()


class VnfInstanceState(BaseTackerEnum):
    INSTANTIATED = 'INSTANTIATED'
    NOT_INSTANTIATED = 'NOT_INSTANTIATED'

    ALL = (INSTANTIATED, NOT_INSTANTIATED)


class VnfInstanceStateField(BaseEnumField):
    AUTO_TYPE = VnfInstanceState()


class VnfInstanceTaskState(BaseTackerEnum):
    INSTANTIATING = 'INSTANTIATING'
    HEALING = 'HEALING'
    TERMINATING = 'TERMINATING'
    SCALING = 'SCALING'
    ERROR = 'ERROR'

    ALL = (INSTANTIATING, HEALING, TERMINATING, SCALING, ERROR)


class VnfInstanceTaskStateField(BaseEnumField):
    AUTO_TYPE = VnfInstanceTaskState()


class VnfOperationalStateType(BaseTackerEnum):
    STARTED = 'STARTED'
    STOPPED = 'STOPPED'

    ALL = (STARTED, STOPPED)


class VnfOperationalStateTypeField(BaseEnumField):
    AUTO_TYPE = VnfOperationalStateType()


class IpAddressType(BaseTackerEnum):
    IPV4 = 'IPV4'
    IPV6 = 'IPV6'

    ALL = (IPV4, IPV6)


class IpAddressTypeField(BaseEnumField):
    AUTO_TYPE = IpAddressType()


class VnfInstanceTerminationType(BaseTackerEnum):
    FORCEFUL = 'FORCEFUL'
    GRACEFUL = 'GRACEFUL'

    ALL = (FORCEFUL, GRACEFUL)


class VnfInstanceTerminationTypeField(BaseEnumField):
    AUTO_TYPE = VnfInstanceTerminationType()


# SOL003 5.5.4.6 Enumeration: CancelModeType
class VnfInstanceCancelModeType(BaseTackerEnum):
    FORCEFUL = 'FORCEFUL'
    GRACEFUL = 'GRACEFUL'

    ALL = (FORCEFUL, GRACEFUL)


class VnfInstanceCancelTypeField(BaseEnumField):
    AUTO_TYPE = VnfInstanceCancelModeType()


class VnfcState(BaseTackerEnum):
    STARTED = 'STARTED'
    STOPPED = 'STOPPED'

    ALL = (STARTED, STOPPED)


class InstanceOperationalState(BaseTackerEnum):
    STARTING = 'STARTING'
    PROCESSING = 'PROCESSING'
    COMPLETED = 'COMPLETED'
    FAILED_TEMP = 'FAILED_TEMP'
    ROLLING_BACK = 'ROLLING_BACK'
    ROLLED_BACK = 'ROLLED_BACK'

    ALL = (STARTING, PROCESSING, COMPLETED, FAILED_TEMP,
        ROLLING_BACK, ROLLED_BACK)


class InstanceOperationalStateField(BaseEnumField):
    AUTO_TYPE = InstanceOperationalState()


class InstanceOperation(BaseTackerEnum):
    INSTANTIATE = 'INSTANTIATE'
    SCALE = 'SCALE'
    TERMINATE = 'TERMINATE'
    HEAL = 'HEAL'
    MODIFY_INFO = 'MODIFY_INFO'

    ALL = (INSTANTIATE, SCALE,
        TERMINATE, HEAL, MODIFY_INFO)


class InstanceOperationField(BaseEnumField):
    AUTO_TYPE = InstanceOperation()


class LcmOccsOperationState(BaseTackerEnum):
    STARTING = 'STARTING'
    PROCESSING = 'PROCESSING'
    COMPLETED = 'COMPLETED'
    FAILED_TEMP = 'FAILED_TEMP'
    FAILED = 'FAILED'

    ALL = (STARTING, PROCESSING, COMPLETED, FAILED_TEMP, FAILED)


class LcmOccsOperationType(BaseTackerEnum):
    INSTANTIATE = 'INSTANTIATE'
    TERMINATE = 'TERMINATE'
    HEAL = 'HEAL'
    SCALE = 'SCALE'
    CHANGE_EXT_CONN = 'CHANGE_EXT_CONN'

    ALL = (INSTANTIATE, TERMINATE, HEAL, SCALE, CHANGE_EXT_CONN)


class LcmOccsNotificationStatus(BaseTackerEnum):
    START = 'START'
    RESULT = 'RESULT'

    ALL = (START, RESULT)


class ResourceChangeType(BaseTackerEnum):
    ADDED = 'ADDED'
    REMOVED = 'REMOVED'
    MODIFIED = 'MODIFIED'
    TEMPORARY = 'TEMPORARY'

    ALL = (ADDED, REMOVED, MODIFIED, TEMPORARY)


class LcmOccsNotificationType(BaseTackerEnum):
    VNF_OP_OCC_NOTIFICATION = 'VnfLcmOperationOccurrenceNotification'
    VNF_ID_CREATION_NOTIFICATION = 'VnfIdentifierCreationNotification'

    ALL = (VNF_OP_OCC_NOTIFICATION)


class VnfStatus(BaseTackerEnum):
    ACTIVE = 'ACTIVE'
    INACTIVE = 'INACTIVE'

    ALL = (ACTIVE, INACTIVE)


class InstanceOperation(BaseTackerEnum):
    MODIFY_INFO = 'MODIFY_INFO'


class ErrorPoint(BaseTackerEnum):
    INITIAL = 0
    NOTIFY_PROCESSING = 1
    VNF_CONFIG_START = 2
    PRE_VIM_CONTROL = 3
    POST_VIM_CONTROL = 4
    INTERNAL_PROCESSING = 5
    VNF_CONFIG_END = 6
    NOTIFY_COMPLETED = 7



# Copyright (c) 2019 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from tacker.common import exceptions
from tacker import context
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnfSoftwareImages(SqlTestCase):

    def setUp(self):
        super(TestVnfSoftwareImages, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_package = self._create_vnf_package()
        self.vnf_deployment_flavour = self._create_vnf_deployment_flavour()
        self.vnf_softwate_images = self._create_vnf_softwate_images()

    def _create_vnf_package(self):
        vnfpkgm = objects.VnfPackage(context=self.context,
                                     **fakes.vnf_package_data)
        vnfpkgm.create()
        return vnfpkgm

    def _create_vnf_deployment_flavour(self):
        flavour_data = fakes.vnf_deployment_flavour
        flavour_data.update({'package_uuid': self.vnf_package.id})
        vnf_deployment_flavour = objects.VnfDeploymentFlavour(
            context=self.context, **flavour_data)
        vnf_deployment_flavour.create()
        return vnf_deployment_flavour

    def _create_vnf_softwate_images(self):
        software_image = fakes.software_image
        software_image.update(
            {'flavour_uuid': self.vnf_deployment_flavour.id})

        vnf_soft_image_obj = objects.VnfSoftwareImage(
            context=self.context, **software_image)
        vnf_soft_image_obj.create()
        return vnf_soft_image_obj

    def test_create(self):
        software_image = fakes.software_image
        software_image.update(
            {'flavour_uuid': self.vnf_deployment_flavour.id})

        vnf_soft_image_obj = objects.VnfSoftwareImage(
            context=self.context, **software_image)
        vnf_soft_image_obj.create()
        self.assertTrue(vnf_soft_image_obj.id)

    def test_software_image_create_with_id(self):
        software_image = fakes.software_image
        software_image.update({'id': uuidsentinel.id})
        vnf_soft_image_obj = objects.VnfSoftwareImage(
            context=self.context, **software_image)
        self.assertRaises(
            exceptions.ObjectActionError,
            vnf_soft_image_obj.create)

    def test_get_by_id(self):
        vnf_software_images = objects.VnfSoftwareImage.get_by_id(
            self.context, self.vnf_softwate_images.id, expected_attrs=None)
        self.compare_obj(self.vnf_softwate_images, vnf_software_images)

    def test_get_by_id_with_no_existing_id(self):
        self.assertRaises(
            exceptions.VnfSoftwareImageNotFound,
            objects.VnfSoftwareImage.get_by_id, self.context,
            uuidsentinel.invalid_uuid)

    def test_attribute_with_valid_data(self):
        data = {'id': self.vnf_softwate_images.id}
        vnf_software_image_obj = objects.VnfSoftwareImage(
            context=self.context, **data)
        vnf_software_image_obj.obj_load_attr('name')
        self.assertEqual('test', vnf_software_image_obj.name)

    def test_invalid_attribute(self):
        self.assertRaises(exceptions.ObjectActionError,
                          self.vnf_softwate_images.obj_load_attr, 'invalid')

    def test_obj_load_attr_without_context(self):
        data = {'id': self.vnf_softwate_images.id}
        vnf_software_image_obj = objects.VnfSoftwareImage(**data)
        self.assertRaises(exceptions.OrphanedObjectError,
                          vnf_software_image_obj.obj_load_attr, 'name')

    def test_obj_load_attr_without_id_in_object(self):
        data = {'name': self.vnf_softwate_images.name}
        vnf_software_image_obj = objects.VnfSoftwareImage(
            context=self.context, **data)
        self.assertRaises(exceptions.ObjectActionError,
                          vnf_software_image_obj.obj_load_attr, 'name')



# Copyright 2011 OpenStack Foundation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy
from sqlalchemy.orm.properties import RelationshipProperty

from tacker._i18n import _
from tacker.common import exceptions as n_exc


def paginate_query(query, model, limit, sorts, marker_obj=None):
    """Returns a query with sorting / pagination criteria added.

    Pagination works by requiring a unique sort key, specified by sorts.
    (If sort keys is not unique, then we risk looping through values.)
    We use the last row in the previous page as the 'marker' for pagination.
    So we must return values that follow the passed marker in the order.
    With a single-valued sort key, this would be easy: sort_key > X.
    With a compound-values sort key, (k1, k2, k3) we must do this to repeat
    the lexicographical ordering:
    (k1 > X1) or (k1 == X1 && k2 > X2) or (k1 == X1 && k2 == X2 && k3 > X3)
    The reason of didn't use OFFSET clause was it don't scale, please refer
    discussion at https://lists.launchpad.net/openstack/msg02547.html

    We also have to cope with different sort directions.

    Typically, the id of the last row is used as the client-facing pagination
    marker, then the actual marker object must be fetched from the db and
    passed in to us as marker.

    :param query: the query object to which we should add paging/sorting
    :param model: the ORM model class
    :param limit: maximum number of items to return
    :param sorts: array of attributes and direction by which results should
                 be sorted
    :param marker: the last item of the previous page; we returns the next
                    results after this value.
    :rtype: sqlalchemy.orm.query.Query
    :return: The query with sorting/pagination added.
    """
    if not sorts:
        return query

    # A primary key must be specified in sort keys
    assert not (limit and
                len(set(dict(sorts).keys()) &
                    set(model.__table__.primary_key.columns.keys())) == 0)

    # Add sorting
    for sort_key, sort_direction in sorts:
        sort_dir_func = sqlalchemy.asc if sort_direction else sqlalchemy.desc
        try:
            sort_key_attr = getattr(model, sort_key)
        except AttributeError:
            # Extension attribute doesn't support for sorting. Because it
            # existed in attr_info, it will be catched at here
            msg = _("%s is invalid attribute for sort_key") % sort_key
            raise n_exc.BadRequest(resource=model.__tablename__, msg=msg)
        if isinstance(sort_key_attr.property, RelationshipProperty):
            msg = _("The attribute '%(attr)s' is reference to other "
                    "resource, can't used by sort "
                    "'%(resource)s'") % {'attr': sort_key,
                                         'resource': model.__tablename__}
            raise n_exc.BadRequest(resource=model.__tablename__, msg=msg)
        query = query.order_by(sort_dir_func(sort_key_attr))

    # Add pagination
    if marker_obj:
        marker_values = [getattr(marker_obj, sort[0]) for sort in sorts]

        # Build up an array of sort criteria as in the docstring
        criteria_list = []
        for i, sort in enumerate(sorts):
            crit_attrs = [(getattr(model, sorts[j][0]) == marker_values[j])
                          for j in range(i)]
            model_attr = getattr(model, sort[0])
            if sort[1]:
                crit_attrs.append((model_attr > marker_values[i]))
            else:
                crit_attrs.append((model_attr < marker_values[i]))

            criteria = sqlalchemy.sql.and_(*crit_attrs)
            criteria_list.append(criteria)

        f = sqlalchemy.sql.or_(*criteria_list)
        query = query.filter(f)

    if limit:
        query = query.limit(limit)

    return query



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import uuid

from oslo_serialization import jsonutils

from sqlalchemy.types import String
from sqlalchemy.types import Text
from sqlalchemy.types import TypeDecorator


class Uuid(TypeDecorator):
    impl = String(36)

    cache_ok = True

    def process_bind_param(self, value, dialect):
        if value is not None:
            # If value is typed as List, only use first element.
            if isinstance(value, list):
                value = value[0]
            try:
                uuid.UUID(value, version=4)
            except ValueError:
                raise ValueError(
                    "Invalid format. It should be in UUID v4 format")

        return value

    def process_result_value(self, value, dialect):
        return value


class Json(TypeDecorator):
    impl = Text

    cache_ok = False

    def process_bind_param(self, value, dialect):
        return jsonutils.dump_as_bytes(value)

    def process_result_value(self, value, dialect):
        if value is None:
            return None
        return jsonutils.loads(value)



# Copyright (C) 2021 Nippon Telegraph and Telephone Corporation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


import functools

from tacker.api.validation import validators
from tacker.common import exceptions as tacker_ex

from tacker.sol_refactored.api import api_version
from tacker.sol_refactored.common import exceptions as sol_ex


class SolSchemaValidator(validators._SchemaValidator):
    def validate(self, *args, **kwargs):
        try:
            super(SolSchemaValidator, self).validate(*args, **kwargs)
        except tacker_ex.ValidationError as ex:
            raise sol_ex.SolValidationError(detail=str(ex))


def schema(request_body_schema, min_version, max_version=None):

    def add_validator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            ver = kwargs['request'].context.api_version
            min_ver = api_version.APIVersion(min_version)
            max_ver = api_version.APIVersion(max_version)
            if ver.matches(min_ver, max_ver):
                if 'body' not in kwargs:
                    raise sol_ex.SolValidationError(detail="body is missing")
                schema_validator = SolSchemaValidator(request_body_schema)
                schema_validator.validate(kwargs['body'])

            return func(*args, **kwargs)
        return wrapper

    return add_validator



# Copyright (c) 2019 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from oslo_utils import uuidutils

from tacker.common import exceptions
from tacker import context
from tacker.db.db_sqlalchemy import models
from tacker import objects
from tacker.tests.unit.db.base import SqlTestCase
from tacker.tests.unit.objects import fakes
from tacker.tests import uuidsentinel


class TestVnfPackage(SqlTestCase):

    def setUp(self):
        super(TestVnfPackage, self).setUp()
        self.context = context.get_admin_context()
        self.vnf_package = self._create_vnf_package()

    def _create_vnf_package(self):
        vnfpkgm = objects.VnfPackage(context=self.context,
                                     **fakes.vnf_package_data)
        vnfpkgm.create()
        return vnfpkgm

    def test_create(self):
        vnfpkgm = objects.VnfPackage(context=self.context,
                                     **fakes.vnf_package_data)
        vnfpkgm.create()
        self.assertTrue(vnfpkgm.id)
        self.assertEqual('CREATED', vnfpkgm.onboarding_state)
        self.assertEqual('NOT_IN_USE', vnfpkgm.usage_state)
        self.assertEqual('DISABLED', vnfpkgm.operational_state)
        self.assertEqual(0, vnfpkgm.size)

    def test_create_without_user_define_data(self):
        vnf_pack = fakes.vnf_package_data
        vnf_pack.update({'user_data': {}})
        vnfpkgm = objects.VnfPackage(context=self.context, **vnf_pack)
        vnfpkgm.create()
        self.assertTrue(vnfpkgm.id)
        self.assertEqual('CREATED', vnfpkgm.onboarding_state)
        self.assertEqual('NOT_IN_USE', vnfpkgm.usage_state)
        self.assertEqual('DISABLED', vnfpkgm.operational_state)

    @mock.patch.object(uuidutils, 'generate_uuid')
    @mock.patch('tacker.objects.vnf_package._vnf_package_create')
    def test_create_ignore_flavours(
            self, mock_vnf_package_create, mock_uuid):
        fake_data = fakes.fake_vnf_package_response
        mock_uuid.return_value = fake_data['id']
        mock_vnf_package_create.return_value = models.VnfPackage(**fake_data)

        flavour_data = fakes.vnf_deployment_flavour
        flavour_data.update({'package_uuid': self.vnf_package.id})
        vnf_deployment_flavour = objects.VnfDeploymentFlavour(**flavour_data)

        fake_vnf_deployment_flavours = objects.VnfDeploymentFlavoursList(
            objects=[vnf_deployment_flavour])
        vnf_pack = fakes.vnf_package_data
        vnf_pack.update(
            {'vnf_deployment_flavours': fake_vnf_deployment_flavours})
        vnfpkgm = objects.VnfPackage(context=self.context, **vnf_pack)
        vnfpkgm.create()

        mock_vnf_package_create.assert_called_once_with(
            self.context, fake_data, user_data={'abc': 'xyz'})

    def test_get_by_id(self):
        vnfpkgm = objects.VnfPackage.get_by_id(self.context,
                                               self.vnf_package.id,
                                               expected_attrs=None)
        self.compare_obj(self.vnf_package, vnfpkgm,
                         allow_missing=['vnf_deployment_flavours',
                                        'vnf_artifacts'])

    def test_get_by_id_with_no_existing_id(self):
        self.assertRaises(
            exceptions.VnfPackageNotFound,
            objects.VnfPackage.get_by_id, self.context,
            uuidsentinel.invalid_uuid)

    def test_create_with_id(self):
        vnfpkgm = objects.VnfPackage(context=self.context,
                **fakes.vnf_package_data)
        vnfpkgm['id'] = uuidsentinel.uuid
        vnfpkgm.create()
        self.assertTrue(vnfpkgm.id)
        self.assertEqual('CREATED', vnfpkgm.onboarding_state)
        self.assertEqual('NOT_IN_USE', vnfpkgm.usage_state)
        self.assertEqual('DISABLED', vnfpkgm.operational_state)
        self.assertEqual(0, vnfpkgm.size)

    def test_save(self):
        self.vnf_package.onboarding_state = 'ONBOARDED'
        self.vnf_package.save()
        self.assertEqual('ONBOARDED', self.vnf_package.onboarding_state)

    def test_save_error(self):
        fake_data = {'id': uuidsentinel.uuid}
        vnf_pack_obj = objects.VnfPackage(context=self.context,
                                          **fake_data)
        self.assertRaises(exceptions.VnfPackageNotFound, vnf_pack_obj.save)

    @mock.patch('tacker.objects.vnf_package._destroy_vnf_package')
    def test_destroy(self, mock_vnf_destroy):
        self.vnf_package.destroy(self.context)
        mock_vnf_destroy.assert_called_with(self.context, self.vnf_package.id)

    def test_destroy_without_id(self):
        vnf_obj = objects.VnfPackage(context=self.context)
        self.assertRaises(exceptions.ObjectActionError, vnf_obj.destroy,
                          self.context)

    def test_get_all(self):
        result = objects.VnfPackagesList.get_all(self.context,
                                                 expected_attrs=None)
        self.assertIsInstance(result.objects, list)
        self.assertTrue(result.objects)

    def test_get_by_id_with_flavours(self):
        flavour_data = fakes.vnf_deployment_flavour
        flavour_data.update({'package_uuid': self.vnf_package.id})
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(
            context=self.context, **flavour_data)
        vnf_deployment_flavour_obj.create()
        vnfpkgm = objects.VnfPackage.get_by_id(
            self.context, self.vnf_package.id,
            expected_attrs=['vnf_deployment_flavours'])
        self.assertEqual(1, len(vnfpkgm.vnf_deployment_flavours.objects))
        self.compare_obj(vnf_deployment_flavour_obj,
                         vnfpkgm.vnf_deployment_flavours[0],
                         allow_missing=['software_images',
                                        'updated_at',
                                        'deleted', 'deleted_at'])

    def test_get_by_id_without_flavours(self):
        vnfpkgm = objects.VnfPackage.get_by_id(
            self.context, self.vnf_package.id, expected_attrs=None)
        self.assertEqual([], vnfpkgm.vnf_deployment_flavours.objects)

    def test_attribute_with_valid_data(self):
        data = {'id': self.vnf_package.id}
        vnf_pack_obj = objects.VnfPackage(context=self.context, **data)
        vnf_pack_obj.obj_load_attr('user_data')
        self.assertEqual({'abc': 'xyz'}, vnf_pack_obj.user_data)

    def test_invalid_attribute(self):
        self.assertRaises(exceptions.ObjectActionError,
                          self.vnf_package.obj_load_attr, 'invalid')

    def test_obj_load_attr_without_context(self):
        data = {'id': self.vnf_package.id}
        vnf_package_obj = objects.VnfPackage(**data)
        self.assertRaises(exceptions.OrphanedObjectError,
                          vnf_package_obj.obj_load_attr, 'algorithm')

    def test_obj_load_attr_without_id_in_object(self):
        data = {'user_data': {'tests': 'test_data'}}
        vnf_deployment_flavour_obj = objects.VnfDeploymentFlavour(
            context=self.context, **data)
        self.assertRaises(
            exceptions.ObjectActionError,
            vnf_deployment_flavour_obj.obj_load_attr, 'algorithm')

    def test_vnf_package_list_by_filter(self):
        filters = {'field': 'onboarding_state', 'model': 'VnfPackage',
                   'value': 'CREATED',
                   'op': '=='}
        vnfpkgm_list = objects.VnfPackagesList.get_by_filters(
            self.context, filters=filters)
        self.assertEqual(1, len(vnfpkgm_list))

    def test_obj_make_compatible(self):
        data = {'id': self.vnf_package.id}
        vnf_package_obj = objects.VnfPackage(context=self.context, **data)
        fake_vnf_package_obj = objects.VnfPackage(context=self.context, **data)
        obj_primitive = fake_vnf_package_obj.obj_to_primitive('1.0')
        obj = vnf_package_obj.obj_from_primitive(obj_primitive)
        self.assertIn('size', obj.fields)



# Copyright (C) 2019 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
import hashlib
import os
import re
import shutil
from urllib.parse import urlparse
from urllib import request as urllib2
import yaml
import zipfile

from oslo_log import log as logging
from oslo_utils import encodeutils
from oslo_utils import excutils
from tacker.common import utils
from toscaparser.prereq.csar import CSAR
from toscaparser.tosca_template import ToscaTemplate

from tacker.common import exceptions
import tacker.conf
from tacker.extensions import vnfm


HASH_DICT = {
    'sha-224': hashlib.sha224,
    'sha-256': hashlib.sha256,
    'sha-384': hashlib.sha384,
    'sha-512': hashlib.sha512
}
CONF = tacker.conf.CONF
LOG = logging.getLogger(__name__)
TOSCA_META = 'TOSCA-Metadata/TOSCA.meta'
ARTIFACT_KEYS = ['Source', 'Algorithm', 'Hash']
IMAGE_FORMAT_LIST = ['raw', 'vhd', 'vhdx', 'vmdk', 'vdi', 'iso', 'ploop',
                   'qcow2', 'aki', 'ari', 'ami', 'img']


def _check_type(custom_def, node_type, type_list):
    for node_data_type, node_data_type_value in custom_def.items():
        if node_data_type == node_type and node_type in type_list:
            return True, node_data_type_value
        for k, v in node_data_type_value.items():
            if k == 'derived_from':
                if v in type_list and node_type == node_data_type:
                    return True, node_data_type_value
    return False, None


def _get_sw_image_artifact(artifacts):
    if not artifacts:
        return

    for artifact_value in artifacts.values():
        if isinstance(artifact_value, dict):
            if artifact_value.get('type') == 'tosca.artifacts.nfv.SwImage':
                return artifact_value
        elif isinstance(artifact_value, str):
            return {'file': artifact_value}


def _update_default_vnfd_data(node_value, node_type_value):
    vnf_properties = node_value['properties']
    type_properties = node_type_value['properties']
    for property_key, property_value in type_properties.items():
        if property_key == 'descriptor_id':
            # if descriptor_id is parameterized, then get the value from the
            # default property and set it in the vnf_properties.
            if vnf_properties and isinstance(
                    vnf_properties.get('descriptor_id'), dict):
                vnf_properties['descriptor_id'] = property_value.get("default")
    return vnf_properties


def _get_vnf_data(nodetemplates):
    type_list = ['tosca.nodes.nfv.VNF']
    for nt in nodetemplates:
        for node_name, node_value in nt.templates.items():
            type_status, node_type_value = _check_type(nt.custom_def,
                                            node_value['type'], type_list)
            if type_status and node_type_value:
                return _update_default_vnfd_data(node_value, node_type_value)


def _get_instantiation_levels(policies):
    if policies:
        for policy in policies:
            if policy.type_definition.type == \
                    'tosca.policies.nfv.InstantiationLevels':
                return policy.properties


def _update_flavour_data_from_vnf(custom_defs, node_tpl, flavour):
    type_list = ['tosca.nodes.nfv.VNF']

    type_status, _ = _check_type(custom_defs, node_tpl['type'], type_list)
    if type_status and node_tpl['properties']:
        vnf_properties = node_tpl['properties']
        if 'flavour_description' in vnf_properties:
            flavour.update(
                {'flavour_description': vnf_properties[
                    'flavour_description']})

        if 'flavour_id' in vnf_properties and isinstance(
                vnf_properties['flavour_id'], str):
            flavour.update({'flavour_id': vnf_properties['flavour_id']})


def _get_software_image(custom_defs, nodetemplate_name, node_tpl):
    type_list = ['tosca.nodes.nfv.Vdu.Compute',
                 'tosca.nodes.nfv.Vdu.VirtualBlockStorage']
    type_status, _ = _check_type(custom_defs, node_tpl['type'], type_list)
    if type_status:
        properties = node_tpl['properties']
        sw_image_artifact = _get_sw_image_artifact(node_tpl.get('artifacts'))
        if sw_image_artifact:
            image_path = sw_image_artifact['file'].lstrip("./")
            properties['sw_image_data'].update(
                {'software_image_id': nodetemplate_name,
                'image_path': image_path})
            sw_image_data = properties['sw_image_data']
            _convert_software_images_prop_to_fixed_unit(sw_image_data)

            if 'metadata' in sw_image_artifact:
                sw_image_data.update({'metadata':
                    sw_image_artifact['metadata']})
            return sw_image_data


def _convert_software_images_prop_to_fixed_unit(sw_image_data):
    """Update values of 'min_disk', 'min_ram' and 'size' to Bytes.

    Since, the units like MB/MiB/GB/GiB is not stored in the database, we need
    to convert 'min_disk', 'min_ram' and 'size' values to fixed unit before
    saving it in database. Here its converting 'min_disk', 'min_ram', and
    'size' values to Bytes.
    """
    for attribute in ['min_disk', 'min_ram', 'size']:
        if sw_image_data.get(attribute):
            updated_value = utils.MemoryUnit.convert_unit_size_to_num(
                sw_image_data.get(attribute),
                unit='B')
            sw_image_data[attribute] = updated_value


def _populate_flavour_data(tosca):
    flavours = []
    if tosca.nested_tosca_templates_with_topology:
        for tp in tosca.nested_tosca_templates_with_topology:
            sw_image_list = []

            # Setting up flavour data
            flavour_id = tp.substitution_mappings.properties.get('flavour_id')
            if flavour_id:
                flavour = {'flavour_id': flavour_id}
                tpl_dict = dict()

                # get from top-vnfd data
                for key, value in tosca.tpl.items():
                    if key in CONF.vnf_package.get_top_list:
                        tpl_dict[key] = value

                # get from lower-vnfd data
                tpl_dict['topology_template'] = dict()
                tpl_dict['topology_template']['groups'] = \
                    tp.tpl.get('groups')
                tpl_dict['topology_template']['policies'] = \
                    tp.tpl.get('policies')
                tpl_dict['topology_template']['node_templates'] = \
                    deepcopy(tp.tpl.get('node_templates'))
                for e_node in CONF.vnf_package.exclude_node:
                    if tpl_dict['topology_template']['node_templates'].\
                            get(e_node):
                        del (tpl_dict['topology_template']
                            ['node_templates'][e_node])
                tpl_dict['topology_template']['inputs'] = \
                    deepcopy(tp.tpl.get('inputs'))
                for del_input in CONF.vnf_package.del_input_list:
                    if tpl_dict['topology_template']['inputs'].get(del_input):
                        del tpl_dict['topology_template']['inputs'][del_input]
                if len(tpl_dict['topology_template']['inputs']) < 1:
                    del tpl_dict['topology_template']['inputs']

                flavour.update({'tpl_dict': tpl_dict})

                instantiation_levels = _get_instantiation_levels(tp.policies)
                if instantiation_levels:
                    flavour.update(
                        {'instantiation_levels': instantiation_levels})

                mgmt_driver = None
                for template_name, node_tpl in \
                        tp.tpl.get('node_templates').items():
                    # check the flavour property in vnf data
                    _update_flavour_data_from_vnf(
                        tp.custom_defs, node_tpl, flavour)
                    if node_tpl['type'] in CONF.vnf_package.get_lower_list:
                        if node_tpl['type'] == "tosca.nodes.nfv.VDU.Tacker":
                            # get mgmt_driver
                            mgmt_driver_flavour = \
                                node_tpl['properties'].get('mgmt_driver')
                            if mgmt_driver_flavour:
                                if mgmt_driver and \
                                        mgmt_driver_flavour != mgmt_driver:
                                    raise vnfm.MultipleMGMTDriversSpecified()
                                mgmt_driver = mgmt_driver_flavour
                                flavour.update({'mgmt_driver': mgmt_driver})

                for template_name, node_tpl in \
                        tp.tpl.get('node_templates').items():
                    # Update the software image data
                    sw_image = _get_software_image(tp.custom_defs,
                                                   template_name,
                                                   node_tpl)
                    if sw_image:
                        sw_image_list.append(sw_image)

                # Add software images for flavour
                if sw_image_list:
                    flavour.update({'sw_images': sw_image_list})

                if flavour:
                    flavours.append(flavour)

    else:
        _get_flavour_data(tosca.topology_template, flavours)

    return flavours


def _get_flavour_data(tp, flavours):
    sw_image_list = []

    # Setting up flavour data
    flavour_id = tp.substitution_mappings.properties.get('flavour_id')
    if flavour_id:
        if isinstance(flavour_id, dict):
            error_msg = "flavour_id should be string and given" \
                        " {}".format(flavour_id)
            raise exceptions.InvalidCSAR(error_msg)
        flavour = {'flavour_id': flavour_id}
    else:
        flavour = {}
    instantiation_levels = _get_instantiation_levels(tp.policies)
    if instantiation_levels:
        flavour.update({'instantiation_levels': instantiation_levels})
    for template_name, node_tpl in tp.tpl.get('node_templates').items():
        # check the flavour property in vnf data
        _update_flavour_data_from_vnf(tp.custom_defs, node_tpl, flavour)

        # Update the software image data
        sw_image = _get_software_image(tp.custom_defs, template_name,
                                       node_tpl)
        if sw_image:
            sw_image_list.append(sw_image)

    # Add software images for flavour
    if sw_image_list:
        flavour.update({'sw_images': sw_image_list})

    if flavour:
        flavours.append(flavour)


def _get_instantiation_levels_from_policy(tpl_policies):
    """Get defined instantiation levels

    Getting instantiation levels defined under policy type
    'tosca.policies.nfv.InstantiationLevels'.
    """

    levels = []
    for policy in tpl_policies:
        for key, value in policy.items():
            if value.get('type') == 'tosca.policies.nfv.InstantiationLevels'\
                    and value.get('properties', {}).get('levels', {}):
                levels = value.get('properties').get('levels').keys()
                default_level = value.get(
                    'properties').get('default_level')

                if default_level and default_level not in levels:
                    error_msg = "Level {} not found in defined levels" \
                                " {}".format(default_level,
                                             ",".join(sorted(levels)))
                    raise exceptions.InvalidCSAR(error_msg)
    return levels


def _validate_instantiation_levels(policy, instantiation_levels):
    expected_policy_type = ['tosca.policies.nfv.VduInstantiationLevels',
                            'tosca.policies.nfv.'
                            'VirtualLinkInstantiationLevels']
    for policy_name, policy_tpl in policy.items():
        if policy_tpl.get('type') not in expected_policy_type:
            return

        if not instantiation_levels:
            msg = ('Policy of type'
                   ' "tosca.policies.nfv.InstantiationLevels is not defined.')
            raise exceptions.InvalidCSAR(msg)
        if policy_tpl.get('properties'):
            levels_in_policy = policy_tpl.get(
                'properties').get('levels')

            if levels_in_policy:
                invalid_levels = set(levels_in_policy.keys()) - set(
                    instantiation_levels)
            else:
                invalid_levels = set()

            if invalid_levels:
                error_msg = "Level(s) {} not found in defined levels" \
                            " {}".format(",".join(sorted(invalid_levels)),
                                         ",".join(sorted(instantiation_levels)
                                                  ))
                raise exceptions.InvalidCSAR(error_msg)


def _validate_sw_image_data_for_artifact(node_tpl, template_name):
    artifact_names = []
    artifacts = node_tpl.get('artifacts')
    if not artifacts:
        return

    for key, value in artifacts.items():
        if isinstance(value, dict):
            if value.get('type') == 'tosca.artifacts.nfv.SwImage':
                artifact_names.append(key)
        elif isinstance(value, str):
            artifact_names.append(key)

    if len(artifact_names) > 1:
        error_msg = ('artifacts of type "tosca.artifacts.nfv.SwImage"'
                     ' is added more than one time for'
                     ' node %(node)s.') % {'node': template_name}
        raise exceptions.InvalidCSAR(error_msg)

    if artifact_names and node_tpl.get('properties'):
        if not node_tpl.get('properties').get('sw_image_data'):
            error_msg = ('Node property "sw_image_data" is missing for'
                         ' artifact %(artifact_name)s for node %(node)s.') % {
                'artifact_name': artifact_names[0], 'node': template_name}
            raise exceptions.InvalidCSAR(error_msg)


def _validate_sw_image_data_for_artifacts(tosca):
    for tp in tosca.nested_tosca_templates_with_topology:
        for template_name, node_tpl in tp.tpl.get('node_templates').items():
            _validate_sw_image_data_for_artifact(node_tpl, template_name)

    for template in tosca.nodetemplates:
        _validate_sw_image_data_for_artifact(
            template.entity_tpl, template.name)


def _get_data_from_csar(tosca, context, id):
    for tp in tosca.nested_tosca_templates_with_topology:
        policies = tp.tpl.get("policies")
        if policies:
            levels = _get_instantiation_levels_from_policy(policies)
            for policy_tpl in policies:
                _validate_instantiation_levels(policy_tpl, levels)

    _validate_sw_image_data_for_artifacts(tosca)
    vnf_data = _get_vnf_data(tosca.nodetemplates)
    if not vnf_data:
        error_msg = "VNF properties are mandatory"
        raise exceptions.InvalidCSAR(error_msg)

    flavours = _populate_flavour_data(tosca)
    if not flavours:
        error_msg = "No VNF flavours are available"
        raise exceptions.InvalidCSAR(error_msg)

    csar = CSAR(tosca.input_path, tosca.a_file)
    vnf_artifacts = []
    if csar.validate():
        vnf_artifacts = _get_vnf_artifacts(csar)

    return vnf_data, flavours, vnf_artifacts


def _get_vnf_artifacts(csar):
    vnf_artifacts = []
    if csar.is_tosca_metadata:
        if csar._get_metadata("ETSI-Entry-Manifest"):
            manifest_path = csar._get_metadata("ETSI-Entry-Manifest")
            if manifest_path.lower().endswith(".mf"):
                manifest_data = csar.zfile.read(manifest_path)
                vnf_artifacts = _convert_artifacts(
                    vnf_artifacts, manifest_data, csar)
            else:
                invalid_manifest_err_msg = (
                    ('The file "%(manifest)s" in the CSAR "%(csar)s" does not '
                     'contain valid manifest.') %
                    {'manifest': manifest_path, 'csar': csar.path})
                raise exceptions.InvalidCSAR(invalid_manifest_err_msg)
        tosca_data = csar.zfile.read(TOSCA_META)
        vnf_artifacts = _convert_artifacts(vnf_artifacts, tosca_data, csar)

    else:
        filelist = csar.zfile.namelist()
        main_template_file_name = os.path.splitext(
            csar.main_template_file_name)[0]
        for path in filelist:
            if path.lower().endswith(".mf"):
                manifest_file_name = os.path.splitext(path)[0]
                if manifest_file_name == main_template_file_name:
                    manifest_data = csar.zfile.read(path)
                    vnf_artifacts = _convert_artifacts(
                        vnf_artifacts, manifest_data, csar)
                else:
                    invalid_manifest_err_msg = \
                        (('The filename "%(manifest)s" is an invalid name.'
                          'The name must be the same as the main template '
                          'file name.') %
                        {'manifest': path})
                    raise exceptions.InvalidCSAR(invalid_manifest_err_msg)
    # Deduplication
    vnf_artifacts = [dict(t) for t in set([tuple(d.items())
                          for d in vnf_artifacts])]
    return vnf_artifacts


def _convert_artifacts(vnf_artifacts, artifacts_data, csar):
    artifacts_data_split = re.split(b'\n\n+', artifacts_data)

    for data in artifacts_data_split:
        if re.findall(b'.?Name:.?|.?Source:.?|', data):
            # validate key's existence
            if re.findall(b'.?Algorithm:.?|.?Hash:.?', data):
                artifact_data_dict = yaml.safe_load(data)
                if 'Name' in artifact_data_dict.keys():
                    artifact_data_dict.update(
                        {"Source": artifact_data_dict.pop("Name")})
                if 'Content-Type' in artifact_data_dict.keys():
                    del artifact_data_dict['Content-Type']
                if sorted(ARTIFACT_KEYS) != sorted(artifact_data_dict.keys()):
                    missing_key = list(set(ARTIFACT_KEYS) ^
                                       set(artifact_data_dict.keys()))
                    missing_key = sorted(missing_key)
                    invalid_artifact_err_msg = \
                        (('One of the artifact information '
                          'may not have the key("%(key)s")') %
                         {'key': missing_key})
                    raise exceptions.InvalidCSAR(invalid_artifact_err_msg)
                # validate value's existence
                for key, value in artifact_data_dict.items():
                    if not value:
                        invalid_artifact_err_msg = \
                            (('One of the artifact information may not have '
                              'the key value("%(key)s")') % {'key': key})
                        raise exceptions.InvalidCSAR(invalid_artifact_err_msg)
                artifact_path = artifact_data_dict.get('Source')
                if os.path.splitext(artifact_path)[-1][1:] \
                        in IMAGE_FORMAT_LIST:
                    continue
                else:
                    algorithm = artifact_data_dict.get('Algorithm')
                    hash_code = artifact_data_dict.get('Hash')
                    result = _validate_hash(algorithm, hash_code,
                                            csar, artifact_path)
                    if result:
                        vnf_artifacts.append(artifact_data_dict)
                    else:
                        invalid_artifact_err_msg = \
                            (('The hash "%(hash)s" of artifact file '
                              '"%(artifact)s" is an invalid value.') %
                             {'hash': hash_code, 'artifact': artifact_path})
                        raise exceptions.InvalidCSAR(invalid_artifact_err_msg)

    return vnf_artifacts


def _validate_hash(algorithm, hash_code, csar, artifact_path):
    z = zipfile.ZipFile(csar.path)
    algorithm = algorithm.lower()

    # validate Algorithm's value
    if algorithm in HASH_DICT.keys():
        hash_obj = HASH_DICT[algorithm]()
    else:
        invalid_artifact_err_msg = (('The algorithm("%(algorithm)s") of '
                                     'artifact("%(artifact_path)s") is '
                                     'an invalid value.') %
                                    {'algorithm': algorithm,
                                     'artifact_path': artifact_path})
        raise exceptions.InvalidCSAR(invalid_artifact_err_msg)
    filelist = csar.zfile.namelist()

    # validate Source's value
    if artifact_path in filelist:
        hash_obj.update(z.read(artifact_path))
    elif ((urlparse(artifact_path).scheme == 'file') or
          (bool(urlparse(artifact_path).scheme) and
           bool(urlparse(artifact_path).netloc))):
        hash_obj.update(urllib2.urlopen(artifact_path).read())
    else:
        invalid_artifact_err_msg = (('The path("%(artifact_path)s") of '
                                     'artifact Source is an invalid value.') %
                                    {'artifact_path': artifact_path})
        raise exceptions.InvalidCSAR(invalid_artifact_err_msg)

    # validate Hash's value
    if hash_code == hash_obj.hexdigest():
        return True
    else:
        return False


def extract_csar_zip_file(file_path, extract_path):
    try:
        with PreserveZipFilePermissions(file_path, 'r') as zf:
            zf.extractall(extract_path)
    except (RuntimeError, zipfile.BadZipfile) as exp:
        with excutils.save_and_reraise_exception():
            LOG.error("Error encountered while extracting "
                      "csar zip file %(path)s. Error: %(error)s.",
                      {'path': file_path,
                      'error': encodeutils.exception_to_unicode(exp)})
            exp.reraise = False
            raise exceptions.InvalidZipFile(path=file_path)


def load_csar_data(context, package_uuid, zip_path):

    extract_zip_path = os.path.join(CONF.vnf_package.vnf_package_csar_path,
                                    package_uuid)
    extract_csar_zip_file(zip_path, extract_zip_path)

    try:
        tosca = ToscaTemplate(zip_path, None, True)
        return _get_data_from_csar(tosca, context, package_uuid)
    except exceptions.InvalidCSAR as exp:
        with excutils.save_and_reraise_exception():
            LOG.error("Error processing CSAR file %(path)s for vnf package"
                      " %(uuid)s: Error: %(error)s. ",
                      {'path': zip_path, 'uuid': package_uuid,
                    'error': encodeutils.exception_to_unicode(exp)})
    except Exception as exp:
        with excutils.save_and_reraise_exception():
            LOG.error("Tosca parser failed for vnf package %(uuid)s: "
                      "Error: %(error)s. ", {'uuid': package_uuid,
                      'error': encodeutils.exception_to_unicode(exp)})
            exp.reraise = False
            raise exceptions.InvalidCSAR(encodeutils.exception_to_unicode
                                         (exp))


def delete_csar_data(package_uuid):
    # Remove zip and folder from the vnf_package_csar_path
    csar_zip_temp_path = os.path.join(CONF.vnf_package.vnf_package_csar_path,
                                      package_uuid)
    csar_path = os.path.join(CONF.vnf_package.vnf_package_csar_path,
                 package_uuid + ".zip")

    try:
        shutil.rmtree(csar_zip_temp_path)
        os.remove(csar_path)
    except OSError as exc:
        exc_message = encodeutils.exception_to_unicode(exc)
        msg = _('Failed to delete csar folder: '
                '%(csar_path)s, Error: %(exc)s')
        LOG.warning(msg, {'csar_path': csar_path, 'exc': exc_message})


class PreserveZipFilePermissions(zipfile.ZipFile):
    """Patched _extract_member function of zipFile.

    zipfile.ZipFile.extractall function internally calls
    _extract_member function.
    Here _extract_member function is patched to retain the
    file permissions using member.external_attr >> 16.

    Note: First 16 bits of external_attr store permission details.
    """
    def _extract_member(self, member, targetpath, pwd):
        if not isinstance(member, zipfile.ZipInfo):
            member = self.getinfo(member)

        targetpath = super()._extract_member(member, targetpath, pwd)

        attr = member.external_attr >> 16
        if attr != 0:
            os.chmod(targetpath, attr)
        return targetpath



# Copyright 2012 Locaweb.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import shlex

from eventlet.green import subprocess
from eventlet import greenthread
from oslo_log import log as logging
from oslo_utils import excutils

from tacker._i18n import _
from tacker.common import utils


LOG = logging.getLogger(__name__)


def create_process(cmd, root_helper=None, addl_env=None,
                   debuglog=True):
    """Create a process object for the given command.

    The return value will be a tuple of the process object and the
    list of command arguments used to create it.
    """
    if root_helper:
        cmd = shlex.split(root_helper) + cmd
    cmd = map(str, cmd)

    if debuglog:
        LOG.debug("Running command: %s", cmd)
    env = os.environ.copy()
    if addl_env:
        env.update(addl_env)

    obj = utils.subprocess_popen(cmd, shell=False,
                                 stdin=subprocess.PIPE,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 env=env)

    return obj, cmd


def execute(cmd, root_helper=None, process_input=None, addl_env=None,
            check_exit_code=True, return_stderr=False, debuglog=True):
    # Note(gongysh) not use log_levels in config file because
    # some other codes that are not in a loop probably need the debug log
    try:
        obj, cmd = create_process(cmd, root_helper=root_helper,
                                  addl_env=addl_env, debuglog=debuglog)
        _stdout, _stderr = (process_input and
                            obj.communicate(process_input) or
                            obj.communicate())
        obj.stdin.close()
        m = _("\nCommand: %(cmd)s\nExit code: %(code)s\nStdout: %(stdout)r\n"
              "Stderr: %(stderr)r") % {'cmd': cmd, 'code': obj.returncode,
                                       'stdout': _stdout, 'stderr': _stderr}
        if obj.returncode:
            LOG.error(m)
            if check_exit_code:
                raise RuntimeError(m)
        elif debuglog:
            LOG.debug(m)
    finally:
        # NOTE(termie): this appears to be necessary to let the subprocess
        #               call clean something up in between calls, without
        #               it two execute calls in a row hangs the second one
        greenthread.sleep(0)

    return return_stderr and (_stdout, _stderr) or _stdout


def find_child_pids(pid):
    """Retrieve a list of the pids of child processes of the given pid."""

    try:
        raw_pids = execute(['ps', '--ppid', pid, '-o', 'pid='])
    except RuntimeError as e:
        # Unexpected errors are the responsibility of the caller
        with excutils.save_and_reraise_exception() as ctxt:
            # Exception has already been logged by execute
            no_children_found = 'Exit code: 1' in str(e)
            if no_children_found:
                ctxt.reraise = False
                return []
    return [x.strip() for x in raw_pids.split('\n') if x.strip()]



# Copyright (C) 2021 FUJITSU
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from kubernetes import client
from unittest import mock


from tacker.tests.unit import base
from tacker.vnfm.infra_drivers.kubernetes.k8s import translate_outputs
from tacker.vnfm.infra_drivers.kubernetes import translate_template


class TestTOSCAToKubernetes(base.TestCase):
    def setUp(self):
        super(TestTOSCAToKubernetes, self).setUp()
        self.vnf = {
            "vnfd": {
                "service_types": [
                    {
                        "service_type": "vnfd",
                        "id": "ca0d8667-ce35-4f7a-9744-ac4bc7d5579d"
                    }
                ],
                "description": "Sample",
                "tenant_id": "689708956a2d4ae0a27120d3aca6a560",
                "created_at": "2016-10-20 07:38:54",
                "updated_at": None,
                "attributes": {
                    "vnfd":
                    "description: "
                    "Demo example\nmetadata: "
                    "{template_name: sample-tosca-vnfd}\n"
                    "topology_template:\n  "
                    "node_templates:\n    CP1:\n      "
                    "properties: {anti_spoofing_protection: "
                    "false, management: true, order: 0}\n      "
                    "requirements:\n      "
                    "- virtualLink: {node: VL1}\n      "
                    "- virtualBinding: {node: VDU1}\n      "
                    "type: tosca.nodes.nfv.CP.Tacker\n    "
                    "VDU1:\n      "
                    "capabilities:\n        "
                    "nfv_compute:\n          "
                    "properties: {disk_size: 1 GB, "
                    "mem_size: 512 MB, num_cpus: 1}\n      "
                    "properties: {mapping_ports: [80:80] , "
                    "vnfcs: {web:{mem_size: 100 MB, "
                    "config: param0:key1}}}\n      "
                    "type: tosca.nodes.nfv.VDU.Tacker\n    "
                    "VL1:\n      properties: {network_name: "
                    "net_mgmt, vendor: Tacker}\n      "
                    "type: tosca.nodes.nfv.VL\ntosca_definitions_version: "
                    "tosca_simple_profile_for_nfv_1_0_0\n"
                },
                "id": "0fb827e7-32b0-4e5b-b300-e1b1dce8a831",
                "name": "vnfd-sample",
                "template_source": "onboarded or inline"
            }
        }

        self.core_v1_api_client = client.CoreV1Api
        self.app_v1_api_client = client.AppsV1Api
        self.scaling_api_client = client.AutoscalingApi
        self.tosca_to_kubernetes_object = translate_template.TOSCAToKubernetes(
            self.vnf, self.core_v1_api_client, self.app_v1_api_client,
            self.scaling_api_client)

    def test_generate_tosca_kube_objects(self):
        result = self.tosca_to_kubernetes_object.generate_tosca_kube_objects()
        self.assertEqual(result[0].name[:8], "svc-VDU1")
        self.assertEqual(result[0].containers[0].name, "web")
        self.assertEqual(result[0].containers[0].mem_size, 100000000)

    @mock.patch.object(translate_outputs.Transformer, 'deploy')
    def test_deploy_kuberentes_objects(self, mock_deploy):
        mock_deploy.return_value = "name, namespace"
        self.tosca_to_kubernetes_object.deploy_kubernetes_objects()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from oslo_log import log as logging

from tacker.plugins.common import constants
from tacker.vnfm.infra_drivers.openstack import heat_client as hc
from tacker.vnfm.policy_actions import abstract_action
from tacker.vnfm import utils as vnfm_utils
from tacker.vnfm import vim_client

LOG = logging.getLogger(__name__)


class VNFActionRespawn(abstract_action.AbstractPolicyAction):
    def get_type(self):
        return 'respawn'

    def get_name(self):
        return 'respawn'

    def get_description(self):
        return 'Tacker VNF respawning policy'

    def execute_action(self, plugin, context, vnf_dict, args):
        vnf_id = vnf_dict['id']
        LOG.info('vnf %s is dead and needs to be respawned', vnf_id)
        attributes = vnf_dict['attributes']
        vim_id = vnf_dict['vim_id']

        def _update_failure_count():
            failure_count = int(attributes.get('failure_count', '0')) + 1
            failure_count_str = str(failure_count)
            LOG.debug("vnf %(vnf_id)s failure count %(failure_count)s",
                      {'vnf_id': vnf_id, 'failure_count': failure_count_str})
            attributes['failure_count'] = failure_count_str
            attributes['dead_instance_id_' + failure_count_str] = vnf_dict[
                'instance_id']

        def _fetch_vim(vim_uuid):
            vim_res = vim_client.VimClient().get_vim(context, vim_uuid)
            return vim_res

        def _delete_heat_stack(vim_auth):
            placement_attr = vnf_dict.get('placement_attr', {})
            region_name = placement_attr.get('region_name')
            heatclient = hc.HeatClient(auth_attr=vim_auth,
                                       region_name=region_name)
            heatclient.delete(vnf_dict['instance_id'])
            LOG.debug("Heat stack %s delete initiated",
                      vnf_dict['instance_id'])
            vnfm_utils.log_events(context, vnf_dict,
                                  constants.RES_EVT_MONITOR,
                                  "ActionRespawnHeat invoked")

        def _respawn_vnf():
            update_vnf_dict = plugin.create_vnf_sync(context, vnf_dict)
            LOG.info('respawned new vnf %s', update_vnf_dict['id'])
            plugin.config_vnf(context, update_vnf_dict)
            return update_vnf_dict

        if plugin._mark_vnf_dead(vnf_dict['id']):
            _update_failure_count()
            vim_res = _fetch_vim(vim_id)
            if vnf_dict['attributes'].get('monitoring_policy'):
                plugin._vnf_monitor.mark_dead(vnf_dict['id'])
                _delete_heat_stack(vim_res['vim_auth'])
                updated_vnf = _respawn_vnf()
                plugin.add_vnf_to_monitor(context, updated_vnf)
                LOG.debug("VNF %s added to monitor thread",
                          updated_vnf['id'])
            if vnf_dict['attributes'].get('alarming_policy'):
                _delete_heat_stack(vim_res['vim_auth'])
                vnf_dict['attributes'].pop('alarming_policy')
                _respawn_vnf()
